/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: WsdlPublish.java,v 1.1 2004/06/11 09:27:47 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.ant.jonasbase;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.objectweb.jonas.ant.JOnASBaseTask;

/**
 * Generates files for WSDL publish
 * @author Florent Benoit
 */
public class WsdlPublish extends JTask implements BaseTaskItf {

    /**
     * Info for the logger
     */
    private static final String INFO = "[WSDL-Publish] ";

    /**
     * Property for WSDL Handler
     */
    private static final String WSDL_HANDLER_PROPERTY = "jonas.service.ws.wsdlhandlers";

    /**
     * Default property for WSDL handlers
     */
    private static final String TOKEN_WSDLHANDLER = WSDL_HANDLER_PROPERTY + " file1";

    /**
     * List of files (WSDL publish)
     */
    private List files = new ArrayList();

    /**
     * List of uddi (WSDL publish)
     */
    private List uddis = new ArrayList();

    /**
     * Add file (wsdl publish)
     * @param file properties file
     */
    public void addConfiguredFile(File file) {
        files.add(file);
    }

    /**
     * Add UDDI (wsdl publish)
     * @param uddi properties file
     */
    public void addConfiguredUddi(Uddi uddi) {
        uddis.add(uddi);
    }

    /**
     * Execute this task
     */
    public void execute() {

        String fileNameList = "";

        java.io.File jonasBaseConfDir = new java.io.File(getDestDir().getPath() + java.io.File.separator + "conf");

        // Write file Handler
        for (Iterator it = files.iterator(); it.hasNext();) {
            File f = (File) it.next();

            String dir = f.getDir();
            String encoding = f.getEncoding();
            String name = f.getName();

            // Add file to handler list
            if (fileNameList.length() == 0) {
                fileNameList = name;
            } else {
                fileNameList += "," + name;
            }

            String fileName = name + ".properties";

            // Build properties file and write it
            Properties props = new Properties();
            props.put("jonas.service.wsdl.class", File.FILEWSDLHANDLER_CLASS);
            props.put("jonas.service.publish.file.directory", dir);
            props.put("jonas.service.publish.file.encoding", encoding);
            java.io.File writeFile = new java.io.File(jonasBaseConfDir, fileName);
            log(INFO + "Generating a WSDL publish file with name '" + name + "', dir '" + dir + "' and encoding '"
                    + encoding + "' in file '" + writeFile + "'...");
            writePropsToFile(INFO, props, writeFile);
        }

        // Write Uddi Handler
        for (Iterator it = uddis.iterator(); it.hasNext();) {
            Uddi uddi = (Uddi) it.next();

            String name = uddi.getName();
            String username = uddi.getUserName();
            String password = uddi.getPassword();
            String orgName = uddi.getOrgName();
            String orgDesc = uddi.getOrgDesc();
            String orgPersonName = uddi.getOrgPersonName();
            String lifecyclemanagerURL = uddi.getLifecyclemanagerURL();
            String queryManagerURL = uddi.getQueryManagerURL();

            // Add file to handler list
            if (fileNameList.length() == 0) {
                fileNameList = name;
            } else {
                fileNameList += "," + name;
            }

            String fileName = name + ".properties";

            // Build properties file and write it
            Properties props = new Properties();
            props.put("jonas.service.wsdl.class", Uddi.REGISTRYWSDLHANDLER_CLASS);
            props.put("jonas.service.publish.uddi.username", username);
            props.put("jonas.service.publish.uddi.password", password);
            props.put("jonas.service.publish.uddi.organization.name", orgName);
            props.put("jonas.service.publish.uddi.organization.desc", orgDesc);
            props.put("jonas.service.publish.uddi.organization.person_name", orgPersonName);
            props.put("javax.xml.registry.lifeCycleManagerURL", lifecyclemanagerURL);
            props.put("javax.xml.registry.queryManagerURL", queryManagerURL);
            java.io.File writeFile = new java.io.File(jonasBaseConfDir, fileName);
            log(INFO + "Generating a WSDL publish UDDI with name '" + name + "' in file '" + writeFile + "'...");
            writePropsToFile(INFO, props, writeFile);
        }

        // Now set the handlers to the existing list for the property
        JReplace propertyReplace = new JReplace();
        propertyReplace.setProject(getProject());
        propertyReplace.setConfigurationFile(JOnASBaseTask.JONAS_CONF_FILE);
        propertyReplace.setDestDir(new java.io.File(getDestDir().getPath()));
        propertyReplace.setToken(TOKEN_WSDLHANDLER);
        propertyReplace.setValue(WSDL_HANDLER_PROPERTY + SEPARATORS + fileNameList);
        log(INFO + "Adding WSDL handlers '" + fileNameList + "' in " + JOnASBaseTask.JONAS_CONF_FILE + " file.");
        propertyReplace.execute();

    }

    /**
     * Defines a file for WSDL publish
     * @author Florent Benoit
     */
    public static class File {

        /**
         * Directory for publish
         */
        private String dir = null;

        /**
         * Class for FileWSDL handler
         */
        public static final String FILEWSDLHANDLER_CLASS = "org.objectweb.jonas.ws.handler.FileWSDLHandler";

        /**
         * Default encoding
         */
        private static final String DEFAULT_ENCODING = "UTF-8";

        /**
         * User defined encoding
         */

        private String encoding = DEFAULT_ENCODING;

        /**
         * Name of the file
         */
        private String name = null;

        /**
         * Gets the directory
         * @return the directory
         */
        public String getDir() {
            return dir;
        }

        /**
         * Sets the directory
         * @param dir the directory to set
         */
        public void setDir(String dir) {
            this.dir = dir;
        }

        /**
         * Gets the encoding
         * @return the encoding.
         */
        public String getEncoding() {
            return encoding;
        }

        /**
         * Sets the encoding
         * @param encoding the encoding to set.
         */
        public void setEncoding(String encoding) {
            this.encoding = encoding;
        }

        /**
         * Gets the name
         * @return the name.
         */
        public String getName() {
            return name;
        }

        /**
         * Sets the name
         * @param name name to set.
         */
        public void setName(String name) {
            this.name = name;
        }
    }

    /**
     * Defines a UDDI configuration file for WSDL publish
     * @author Florent Benoit
     */
    public static class Uddi {

        /**
         * Class for RegistryWsdl handler
         */
        public static final String REGISTRYWSDLHANDLER_CLASS = "org.objectweb.jonas.ws.handler.RegistryWSDLHandler";

        /**
         * Name of the file
         */
        private String name = null;

        /**
         * Username
         */
        private String userName = null;

        /**
         * Password
         */
        private String password = null;

        /**
         * Organization name
         */
        private String orgName = null;

        /**
         * Organization Description
         */
        private String orgDesc = null;

        /**
         * Organization Person name
         */
        private String orgPersonName = null;

        /**
         * url : lifecyclemanagerURL
         */
        private String lifecyclemanagerURL = null;

        /**
         * url : queryManagerURL
         */
        private String queryManagerURL = null;

        /**
         * Gets the name
         * @return the name.
         */
        public String getName() {
            return name;
        }

        /**
         * Sets the name
         * @param name name to set.
         */
        public void setName(String name) {
            this.name = name;
        }

        /**
         * @return Returns the lifecyclemanagerURL.
         */
        public String getLifecyclemanagerURL() {
            return lifecyclemanagerURL;
        }

        /**
         * @param lifecyclemanagerURL The lifecyclemanagerURL to set.
         */
        public void setLifecyclemanagerURL(String lifecyclemanagerURL) {
            this.lifecyclemanagerURL = lifecyclemanagerURL;
        }

        /**
         * @return Returns the orgDesc.
         */
        public String getOrgDesc() {
            return orgDesc;
        }

        /**
         * @param orgDesc The orgDesc to set.
         */
        public void setOrgDesc(String orgDesc) {
            this.orgDesc = orgDesc;
        }

        /**
         * @return Returns the orgName.
         */
        public String getOrgName() {
            return orgName;
        }

        /**
         * @param orgName The orgName to set.
         */
        public void setOrgName(String orgName) {
            this.orgName = orgName;
        }

        /**
         * @return Returns the orgPersonName.
         */
        public String getOrgPersonName() {
            return orgPersonName;
        }

        /**
         * @param orgPersonName The orgPersonName to set.
         */
        public void setOrgPersonName(String orgPersonName) {
            this.orgPersonName = orgPersonName;
        }

        /**
         * @return Returns the password.
         */
        public String getPassword() {
            return password;
        }

        /**
         * @param password The password to set.
         */
        public void setPassword(String password) {
            this.password = password;
        }

        /**
         * @return Returns the queryManagerURL.
         */
        public String getQueryManagerURL() {
            return queryManagerURL;
        }

        /**
         * @param queryManagerURL The queryManagerURL to set.
         */
        public void setQueryManagerURL(String queryManagerURL) {
            this.queryManagerURL = queryManagerURL;
        }

        /**
         * @return Returns the userName.
         */
        public String getUserName() {
            return userName;
        }

        /**
         * @param userName The userName to set.
         */
        public void setUserName(String userName) {
            this.userName = userName;
        }
    }

}