/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: Pool.java,v 1.58 2004/11/17 20:54:07 ehardesty Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.dbm;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import javax.sql.XAConnection;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jdbc_xa.ConnectionImpl;
import org.objectweb.jonas.jdbc_xa.XADataSourceImpl;
import org.objectweb.jonas.jtm.TransactionService;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.transaction.jta.TransactionManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Connection Pool
 * @author Philippe Durieux
 * Contributor(s):
 * 00/01/08 Markus Fritz
 *          make connection pooling thread safe (with same tx)
 *          introduce MaxAge for connections
 *          optimize pooling
 *          add some 'paranoia code': test conns before returning them from the pool.
 * 00/18/04 Jun Inamori (j-office@osa.att.ne.jp)
 *          For closing correctly all connections at server shutdown.
 * 01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism.
 * 02/01/15 Dean Jennings - Map   instead of Hashtable  for lists
 * 03/05/15 Adriana Danes - Introduce pool size configuration
 */
public class Pool {

    // ----------------------------------------------------------------
    // Private data
    // ----------------------------------------------------------------

    /**
     * Logger used for traces
     */
    private static Logger logger = null;

    /**
     * Keep a ref on the Transaction Manager.
     */
    private TransactionManager tm = null;

    /**
     * My PoolKeeper : This thread periodically adjust pooling
     * and compute statistics.
     */
    private PoolKeeper poolKeeper;

    /**
     * This HashMap gives the pool item by its XAConnection.
     * The size of this table gives the current number of opened connection
     */
    private Map xac2item = new HashMap();

    /**
     * List of free available items
     * This avoids closing and reopening physical connections.
     * We try to keep a minimum of minConPool elements here.
     */
    private LinkedList freeList = new LinkedList();

    /**
     * This HashMap gives the pool item by its transaction
     * Requests with same tx get always the same conn
     */
    private Map tx2item = new HashMap();

    /**
     * the pool's connection manager
     */
    private ConnectionManager cmgr;

    /**
     * the XA datasource managed by the connection manager
     */
    private XADataSourceImpl xads;

    /**
     * count max waiters during current period.
     */
    private int waiterCount = 0;

    /**
     * count max waiting time during current period.
     */
    private long waitingTime = 0;

    /**
     * count max busy connection during current period.
     */
    private int busyMax = 0;

    /**
     * count min busy connection during current period.
     */
    private int busyMin = 0;

    // ----------------------------------------------------------------
    // Configuration
    // Each attibute have setter and getter, plus a default value.
    // ----------------------------------------------------------------

    /**
     * High Value for no limit for the connection pool
     */
    private static final int NO_LIMIT = 99999;

    /**
     * Nb of milliseconds in a day
     */
    private static final long ONE_DAY = 1440L * 60L * 1000L;

    /**
     * max number of remove at once in the freelist
     * We avoid removing too much items at once for perf reasons.
     */
    private static final int MAX_REMOVE_FREELIST = 10;

    /**
     * minimum size of the connection pool
     */
    private int poolMin = 0;

    /**
     * @return min pool size
     */
    public int getPoolMin() {
        return poolMin;
    }

    /**
     * @param min minimum connection pool size to be set.
     */
    public synchronized void setPoolMin(int min) {
        if (poolMin != min) {
            poolMin = min;
            adjust();
        }
    }

    /**
     * maximum size of the connection pool.
     * default value is "NO LIMIT".
     */
    private int poolMax = NO_LIMIT;

    /**
     * @return actual max pool size
     */
    public int getPoolMax() {
        return poolMax;
    }

    /**
     * @param max max pool size. -1 means "no limit".
     */
    public synchronized void setPoolMax(int max) {
        if (poolMax != max) {
            if (max < 0 || max > NO_LIMIT) {
                if (currentWaiters > 0) {
                    notify();
                }
                poolMax = NO_LIMIT;
            } else {
                if (currentWaiters > 0 && poolMax < max) {
                    notify();
                }
                poolMax = max;
                adjust();
            }
        }
    }

    /**
     * Max age of a Connection in milliseconds.
     * When the time is elapsed, the connection will be closed.
     * This avoids keeping connections open too long for nothing.
     */
    private long maxAge = ONE_DAY;

    /**
     * Same value in mns
     */
    private int maxAgeMn;

    /**
     * @return max age for connections (in mm)
     */
    public int getMaxAge() {
        return maxAgeMn;
    }

    /**
     * @return max age for connections (in millisec)
     */
    public long getMaxAgeMilli() {
        return maxAge;
    }

    /**
     * @param mn max age of connection in minutes
     */
    public void setMaxAge(int mn) {
        maxAgeMn = mn;
        // set times in milliseconds
        maxAge = mn * 60L * 1000L;
    }

    /**
     * max open time for a connection, in millisec
     */
    private long maxOpenTime = ONE_DAY;

    /**
     * Same value in mn
     */
    private int maxOpenTimeMn;

    /**
     * @return max age for connections (in mns)
     */
    public int getMaxOpenTime() {
        return maxOpenTimeMn;
    }

    /**
     * @return max age for connections (in millisecs)
     */
    public long getMaxOpenTimeMilli() {
        return maxOpenTime;
    }

    /**
     * @param mn max time of open connection in minutes
     */
    public void setMaxOpenTime(int mn) {
        maxOpenTimeMn = mn;
        // set times in milliseconds
        maxOpenTime = mn * 60L * 1000L;
    }

    /**
     * max nb of milliseconds to wait for a connection when pool is full
     */
    private long waiterTimeout = 10000;

    /**
     * @return waiter timeout in seconds
     */
    public int getMaxWaitTime() {
        return (int) (waiterTimeout / 1000L);
    }

    /**
     * @param sec max time to wait for a connection, in seconds
     */
    public void setMaxWaitTime(int sec) {
        waiterTimeout = sec * 1000L;
    }

    /**
     * max nb of waiters allowed to wait for a Connection
     */
    private int maxWaiters = 1000;

    /**
     * @return max nb of waiters
     */
    public int getMaxWaiters() {
        return maxWaiters;
    }

    /**
     * @param nb max nb of waiters
     */
    public void setMaxWaiters(int nb) {
        maxWaiters = nb;
    }

    /**
     * sampling period in sec.
     */
    private int samplingPeriod = 60; //default sampling period

    /**
     * @return sampling period in sec.
     */
    public int getSamplingPeriod() {
        return samplingPeriod;
    }

    /**
     * @param sec sampling period in sec.
     */
    public void setSamplingPeriod(int sec) {
        if (sec > 0) {
            samplingPeriod = sec;
            poolKeeper.setSamplingPeriod(sec);
        }
    }

    /**
     * Level of checking on connections when got from the pool.
     * this avoids reusing bad connections because too old, for example
     * when database was restarted...
     * 0 = no checking
     * 1 = check that still physically opened.
     * 2 = try a null statement.
     */
    private int checkLevel = 1; // default = 1

    /**
     * @return connection checking level
     */
    public int getCheckLevel() {
        return checkLevel;
    }

    /**
     * @param level jdbc connection checking level (0, 1, or 2)
     */
    public void setCheckLevel(int level) {
        checkLevel = level;
    }

    /**
     * test statement used when checkLevel = 2.
     */
    private String testStatement;

    /**
     * @return test statement used when checkLevel = 2.
     */
    public String getTestStatement() {
        return testStatement;
    }

    /**
     * @param s test statement
     */
    public void setTestStatement(String s) {
        testStatement = s;
    }

    // ----------------------------------------------------------------
    // Monitoring Attributes
    // Each attribute should have a get accessor.
    // ----------------------------------------------------------------

    /**
     *  maximum nb of busy connections in last sampling period
     */
    private int busyMaxRecent = 0;

    /**
     * @return maximum nb of busy connections in last sampling period
     */
    public int getBusyMaxRecent() {
        return busyMaxRecent;
    }

    /**
     *  minimum nb of busy connections in last sampling period
     */
    private int busyMinRecent = 0;

    /**
     * @return minimum nb of busy connections in last sampling period
     */
    public int getBusyMinRecent() {
        return busyMinRecent;
    }

    /**
     * nb of threads waiting for a Connection
     */
    private int currentWaiters = 0;

    /**
     * @return current number of connection waiters
     */
    public int getCurrentWaiters() {
        return currentWaiters;
    }

    /**
     * total number of opened physical connections since the datasource creation.
     */
    private int openedCount = 0;

    /**
     * @return int number of physical jdbc connection opened
     */
    public int getOpenedCount() {
        return openedCount;
    }

    /**
     * total nb of physical connection failures
     */
    private int connectionFailures = 0;

    /**
     * @return int number of xa connection failures on open
     */
    public int getConnectionFailures() {
        return connectionFailures;
    }

    /**
     * total nb of connection leaks.
     * A connection leak occurs when the caller never issues a close method
     * on the connection.
     */
    private int connectionLeaks = 0;

    /**
     * @return int number of connection leaks
     */
    public int getConnectionLeaks() {
        return connectionLeaks;
    }

    /**
     * total number of opened connections since the datasource creation.
     */
    private int servedOpen = 0;

    /**
     * @return int number of xa connection served
     */
    public int getServedOpen() {
        return servedOpen;
    }

    /**
     * total nb of open connection failures because waiter overflow
     */
    private int rejectedFull = 0;

    /**
     * @return int number of open calls that were rejected due to waiter overflow
     */
    public int getRejectedFull() {
        return rejectedFull;
    }

    /**
     * total nb of open connection failures because timeout
     */
    private int rejectedTimeout = 0;

    /**
     * @return int number of open calls that were rejected by timeout
     */
    public int getRejectedTimeout() {
        return rejectedTimeout;
    }

    /**
     * total nb of open connection failures for any other reason.
     */
    private int rejectedOther = 0;

    /**
     * @return int number of open calls that were rejected
     */
    public int getRejectedOther() {
        return rejectedOther;
    }

    /**
     * @return int number of open calls that were rejected
     */
    public int getRejectedOpen() {
        return rejectedFull + rejectedTimeout + rejectedOther;
    }

    /**
     * maximum nb of waiters since datasource creation
     */
    private int waitersHigh = 0;

    /**
     * @return maximum nb of waiters since the datasource creation
     */
    public int getWaitersHigh() {
        return waitersHigh;
    }

    /**
     * maximum nb of waiters in last sampling period
     */
    private int waitersHighRecent = 0;

    /**
     * @return maximum nb of waiters in last sampling period
     */
    public int getWaitersHighRecent() {
        return waitersHighRecent;
    }

    /**
     * total nb of waiters since datasource creation
     */
    private int totalWaiterCount = 0;

    /**
     * @return total nb of waiters since the datasource creation
     */
    public int getWaiterCount() {
        return totalWaiterCount;
    }

    /**
     * total waiting time in milliseconds
     */
    private long totalWaitingTime = 0;

    /**
     * @return total waiting time since the datasource creation
     */
    public long getWaitingTime() {
        return totalWaitingTime;
    }

    /**
     * max waiting time in milliseconds
     */
    private long waitingHigh = 0;

    /**
     * @return max waiting time since the datasource creation
     */
    public long getWaitingHigh() {
        return waitingHigh;
    }

    /**
     * max waiting time in milliseconds in last sampling period
     */
    private long waitingHighRecent = 0;

    /**
     * @return max waiting time in last sampling period
     */
    public long getWaitingHighRecent() {
        return waitingHighRecent;
    }

    // ----------------------------------------------------------------
    // Constructors
    // ----------------------------------------------------------------

    /**
     * Pool constructor
     * @param xads XA DataSource implementation created by this pool's Connection manager. Allows
     * for getting XA Connections.
     * @param cmgr the pool's Connection manager has to be registered as ConnectionEventListener of the XA connections
     * created at the pool creation
     * @throws Exception The pool could not be initialized
     */
    public Pool(ConnectionManager cmgr, XADataSourceImpl xads) throws Exception {
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
        this.cmgr = cmgr;
        this.xads = xads;
        try {
            // Modified for use with the gcj compiler
            ServiceManager sm = ServiceManager.getInstance();
            TransactionService ts = (TransactionService) sm.getTransactionService();
            tm = ts.getTransactionManager();
        } catch (Exception e) {
            throw e;
        }
        poolKeeper = new PoolKeeper(this);
        poolKeeper.start();
    }

    // ----------------------------------------------------------------
    // Public Methods
    // ----------------------------------------------------------------

    /**
     * @return int number of xa connection
     */
    public int getCurrentOpened() {
        return xac2item.size();
    }

    /**
     * @return int number of busy xa connection
     */
    public int getCurrentBusy() {
        return xac2item.size() - freeList.size();
    }

    /**
     * compute current min/max busyConnections
     */
    public void recomputeBusy() {
        int busy = getCurrentBusy();
        if (busyMax < busy) {
            busyMax = busy;
        }
        if (busyMin > busy) {
            busyMin = busy;
        }
    }

    /**
     * @return int number of xa connection reserved for tx
     */
    public int getCurrentInTx() {
        return tx2item.size();
    }

    /**
     * lookup connection in the pool for this user/tx
     * @param user user name
     * @param tx Transaction the connection is involved
     * @return the free PoolItem (never null)
     * @throws SQLException Cannot open a connection because the pool's max size is reached
     */
    public synchronized PoolItem openConnection(String user, Transaction tx) throws SQLException {
        PoolItem item = null;

        // If a Connection exists already for this tx, just return it.
        // If no transaction, never reuse a connection already used.
        if (tx != null) {
            item = (PoolItem) tx2item.get(tx);
            if (item != null) {
                logger.log(BasicLevel.DEBUG, "Reuse a Connection for same tx");
                item.open();
                servedOpen++;
                return item;
            }
        }

        // Loop until a valid item is found
        long timetowait = waiterTimeout;
        long starttime = 0;
        while (item == null) {
            // try to find an item in the free list
            if (freeList.isEmpty()) {
                // In case the pool is full, we must wait until a connection is released.
                if (xac2item.size() >= poolMax) {
                    boolean stoplooping = true;
                    // If a timeout has been specified, wait, unless maxWaiters is reached.
                    if (timetowait > 0) {
                        if (currentWaiters < maxWaiters) {
                            currentWaiters++;
                            // Store the maximum concurrent waiters
                            if (waiterCount < currentWaiters) {
                                waiterCount = currentWaiters;
                            }
                            if (starttime == 0) {
                                starttime = System.currentTimeMillis();
                                logger.log(BasicLevel.DEBUG, "Wait for a free Connection" + xac2item.size());
                            }
                            try {
                                wait(timetowait);
                            } catch (InterruptedException ign) {
                                logger.log(BasicLevel.WARN, "Interrupted");
                            } finally {
                                currentWaiters--;
                            }
                            long stoptime = System.currentTimeMillis();
                            long stillwaited = stoptime - starttime;
                            timetowait = waiterTimeout - stillwaited;
                            stoplooping = (timetowait <= 0);
                            if (stoplooping) {
                                // We have been waked up by the timeout.
                                totalWaiterCount++;
                                totalWaitingTime += stillwaited;
                                if (waitingTime < stillwaited) {
                                    waitingTime = stillwaited;
                                }
                            } else {
                                if (!freeList.isEmpty() || xac2item.size() < poolMax) {
                                    // We have been notified by a connection released.
                                    logger.log(BasicLevel.DEBUG, "Notified after " + stillwaited);
                                    totalWaiterCount++;
                                    totalWaitingTime += stillwaited;
                                    if (waitingTime < stillwaited) {
                                        waitingTime = stillwaited;
                                    }
                                }
                                continue;
                            }
                        }
                    }
                    if (stoplooping && freeList.isEmpty() && xac2item.size() >= poolMax) {
                        if (starttime > 0) {
                            rejectedTimeout++;
                            logger.log(BasicLevel.WARN, "Cannot create a Connection - timeout");
                        } else {
                            rejectedFull++;
                            logger.log(BasicLevel.WARN, "Cannot create a Connection");
                        }
                        throw new SQLException("No more connections in " + cmgr.getDatasourceName());
                    }
                    continue;
                }
                logger.log(BasicLevel.DEBUG, "empty free list: Create a new Connection");
                XAConnection xac = null;
                try {
                    // create a new XA Connection
                    xac = xads.getXAConnection();
                    openedCount++;
                } catch (SQLException e) {
                    connectionFailures++;
                    rejectedOther++;
                    logger.log(BasicLevel.WARN, "Cannot create new Connection for tx");
                    throw e;
                }
                item = new PoolItem(this, xac, user);
                // Register the connection manager as a ConnectionEventListener
                xac.addConnectionEventListener(cmgr);
                xac2item.put(xac, item);
            } else {
                item = (PoolItem) freeList.removeFirst();

                // Check the connection before reusing it
                if (checkLevel > 0) {
                    try {
                        ConnectionImpl conn = (ConnectionImpl) item.getXACon().getConnection();
                        if (conn.isPhysicallyClosed()) {
                            logger.log(BasicLevel.WARN, "The JDBC connection has been closed!");
                            destroyItem(item);
                            starttime = 0;
                            item = null;
                            continue;
                        }
                        if (checkLevel > 1) {
                            java.sql.Statement stmt = conn.createStatement();
                            stmt.execute(testStatement);
                            stmt.close();
                        }
                    } catch (Exception e) {
                        logger.log(BasicLevel.ERROR, "DataSource " + cmgr.getDatasourceName()
                                   + " error: removing invalid item:" + e);
                        destroyItem(item);
                        starttime = 0;
                        item = null;
                        continue;
                    }
                }
            }
        }
        recomputeBusy();
        item.setTx(tx);
        if (tx == null) {
            logger.log(BasicLevel.DEBUG, "Got a Connection - no TX: ");
        } else {
            logger.log(BasicLevel.DEBUG, "Got a Connection for TX: ");
            tx2item.put(tx, item);
            // register synchronization
            try {
                tx.registerSynchronization(item);
            } catch (javax.transaction.RollbackException e) {
                /// optimization is probably possible at this point
                logger.log(BasicLevel.WARN, "DataSource " + cmgr.getDatasourceName()
                           + " error: Pool item registered, but tx is rollback only");
            } catch (javax.transaction.SystemException e) {
                logger.log(BasicLevel.ERROR, "DataSource " + cmgr.getDatasourceName()
                           + " error in pool: system exception from transaction manager " + e.getMessage());
            }
        }
        item.open();
        servedOpen++;
        return item;
    }

    /**
     * The transaction has committed (or rolled back). We can return its
     * connections to the pool of available connections.
     * @param tx the non null transaction
     */
    public synchronized void freeConnections(Transaction tx) {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "free connection for Tx = " + tx);
        }
        PoolItem item = (PoolItem) tx2item.remove(tx);
        if (item == null) {
            logger.log(BasicLevel.ERROR, "pool: no connection found to free for Tx = " + tx);
            return;
        }
        item.setTx(null);
        if (item.isOpen()) {
            // Connection may be not closed in case of stateful session bean that
            // keeps connection to be reused in another method of the bean.
            logger.log(BasicLevel.WARN, "WARNING: Connection not closed by caller");
            return;
        }
        freeItem(item);
    }

    /**
     * Mark a specific Connection in the pool as closed.
     * If it is no longer associated to a Tx, we can free it.
     * @param xac XAConnection being closed
     * @param flag TMSUCCESS (normal close) or TMFAIL (error)
     * @return the PoolItem the connection is in (=> tx)
     * or null if error.
     */
    public synchronized PoolItem closeConnection(XAConnection xac, int flag) {
        PoolItem item = (PoolItem) xac2item.get(xac);
        if (item == null) {
            logger.log(BasicLevel.ERROR, "Pool: no item for this xac!");
            return null;
        }
        // The connection will be available only if not associated
        // to a transaction. Else, it will be reusable only for the
        // same transaction.
        if (!item.close()) {
            return null;
        }
        if (item.getTx() != null) {
            logger.log(BasicLevel.DEBUG, "keep the Connection for this tx");
        } else {
            freeItem(item);
        }

        // delist Resource if in transaction
        Transaction tx = null;
        try {
            tx = tm.getTransaction();
        } catch (NullPointerException n) {
            // current is null: we are not in JOnAS Server.
            logger.log(BasicLevel.ERROR, "Pool: should not be used outside a JOnAS Server");
        } catch (SystemException e) {
            logger.log(BasicLevel.ERROR, "Pool: getTransaction failed:" + e);
        }
        if (tx != null && item.isClosed()) {
            try {
                tx.delistResource(xac.getXAResource(), flag);
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Pool: Exception while delisting resource:" + e);
            }
        }

        return item;
    }

    /**
     * make samples with some monitoring values
     */
    public synchronized void sampling() {
        waitingHighRecent = waitingTime;
        if (waitingHigh < waitingTime) {
            waitingHigh = waitingTime;
        }
        waitingTime = 0;

        waitersHighRecent = waiterCount;
        if (waitersHigh < waiterCount) {
            waitersHigh = waiterCount;
        }
        waiterCount = 0;

        busyMaxRecent = busyMax;
        busyMax = getCurrentBusy();
        busyMinRecent = busyMin;
        busyMin = getCurrentBusy();
    }

    /**
     * Adjust the pool size, according to poolMax and poolMin values.
     * Also remove old connections in the freeList.
     */
    public synchronized void adjust() {
        logger.log(BasicLevel.DEBUG, "");

        // Remove max aged elements in freelist
        // - Not more than MAX_REMOVE_FREELIST
        // - Don't reduce pool size less than poolMin
        int count = xac2item.size() - poolMin;
        if (count > 0) {
            if (count > MAX_REMOVE_FREELIST) {
                count = MAX_REMOVE_FREELIST;
            }
            for (Iterator i = freeList.iterator(); i.hasNext();) {
                PoolItem item = (PoolItem) i.next();
                if (item.isAged()) {
                    logger.log(BasicLevel.DEBUG, "remove a timed out connection");
                    i.remove();
                    destroyItem(item);
                    count--;
                    if (count <= 0) {
                        break;
                    }
                }
            }
        }
        recomputeBusy();

        // Close (physically) connections lost (opened for too long time)
        for (Iterator i = xac2item.values().iterator(); i.hasNext();) {
            PoolItem item = (PoolItem) i.next();
            if (item.inactive()) {
                logger.log(BasicLevel.WARN, "close a timed out open connection");
                i.remove();
                // destroy item
                item.remove();
                connectionLeaks++;
                // Notify 1 thread waiting for a Connection.
                if (currentWaiters > 0) {
                    notify();
                }
            }
        }

        // Shrink the pool in case of max pool size
        // This occurs when max pool size has been reduced by jonas admin.
        if (poolMax != NO_LIMIT) {
            while (freeList.size() > poolMin && xac2item.size() > poolMax) {
                PoolItem item = (PoolItem) freeList.removeFirst();
                destroyItem(item);
            }
        }
        recomputeBusy();

        // Recreate more Connections while poolMin is not reached
        while (xac2item.size() < poolMin) {
            XAConnection xac = null;
            try {
                xac = xads.getXAConnection();
                openedCount++;
            } catch (SQLException e) {
                throw new ServiceException("Could not create " + poolMin + " items in the pool : " + e.toString());
            }
            // tx = null. Assumes maxage already configured.
            logger.log(BasicLevel.DEBUG, "Create a new available Connection");
            PoolItem item = new PoolItem(this, xac, null);
            freeList.addLast(item);
            xac2item.put(xac, item);
            xac.addConnectionEventListener(cmgr);
        }
    }

    /**
     * Close all connections in the pool, when server is shut down.
     */
    public synchronized void closeAllConnections() {
        logger.log(BasicLevel.DEBUG, "");

        // Stop the pool keeper, since all connections will be closed.
        poolKeeper.stop();

        // Close physically all connections
        Iterator it = xac2item.keySet().iterator();
        try {
            while (it.hasNext()) {
                XAConnection xac = (XAConnection) it.next();
                xac.close();
            }
        } catch (java.sql.SQLException e) {
            logger.log(BasicLevel.ERROR, "Error while closing a Connection:" + e);
        }
    }

    // ----------------------------------------------------------------
    // Private Methods
    // ----------------------------------------------------------------

    /**
     * Free item and return it in the free list.
     * @param item The item to be freed
     */
    private void freeItem(PoolItem item) {
        // Add it to the free list
        // Even if maxage is reached, because we avoids going under min pool size.
        // PoolKeeper will manage aged connections.
        freeList.addLast(item);
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "item added to freeList: " + freeList.size());
        }
        // Notify 1 thread waiting for a Connection.
        if (currentWaiters > 0) {
            notify();
        }
        recomputeBusy();
    }

    /**
     * Destroy an item because connection closed or error occured
     * @param item The item to be destroyed
     */
    private void destroyItem(PoolItem item) {
        logger.log(BasicLevel.DEBUG, "remove an item in xac2item");
        xac2item.remove(item.getXACon());
        item.remove();
        // Notify 1 thread waiting for a Connection.
        if (currentWaiters > 0) {
            notify();
        }
        recomputeBusy();
    }

}
