/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 * 01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism to remove ThreadData
 *          dependency.
 *
 * --------------------------------------------------------------------------
 * $Id: PoolItem.java,v 1.18 2004/09/07 08:15:22 danesa Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import javax.sql.XAConnection;
import javax.transaction.Synchronization;
import javax.transaction.Transaction;

import org.objectweb.jonas.common.Log;
import org.objectweb.transaction.jta.ResourceManagerEvent;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class represents a connection stored in the pool.
 * It may be in different states:
 * - available and reuseable for this user (open=0, tx=null)
 * - reusable for the same transaction, not already committed (open=?, tx!=null)
 * - still open, but not involved in a Tx (open=1, tx=null)
 * These 2 cases must be considered: close after commit, commit after close.
 * We cannot suppose what shema the user will use and we must handle both correctly.
 */
public class PoolItem implements ResourceManagerEvent, Synchronization {

    /**
     * Count of items. Used for debugging
     */
    private static int count = 0;

    private XAConnection xaConn; // the pooled connection
    private String user;         // the user name
    private int open;            // >0 if connection is open
    private Transaction tx;  // transaction the connection is involved with
    private Transaction enlistedInTx; /// current transaction in case of late enlistment
    private boolean rme = false; // true if registered as ResourceManagerEvent
    private long deathTime;     // When it is expected to die (for closed connections)
    private long closeTime;     // When it is expected to be closed (for opened connections)
    private int number;
    private Pool pool;
    private Logger logger = null;

    /*
     * constructor used by newConnection()
     */
    public PoolItem(Pool pool, XAConnection xac, String user) {
        this.pool = pool;
        this.xaConn = xac;
        this.user = user;
        open = 0;
        deathTime = System.currentTimeMillis() + pool.getMaxAgeMilli();
        number = count++;
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
    }

    /*
     * string form for debugging
     */
    public String toString() {
        Integer i = new Integer(number);
        return i.toString();
    }

    /**
     *  implementation of resource manager event
     */
    public void enlistConnection(Transaction transaction) throws javax.transaction.SystemException  {
        try {
            if (rme) {
                xaConn.getConnection().setAutoCommit(false);
                // the tx value of the pool item does not match with the enlisted value !?
                enlistedInTx = transaction;
                // enlist the resource
                if (logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "enlist XAResource on " + transaction);
                transaction.enlistResource(xaConn.getXAResource());
            }
        } catch (javax.transaction.RollbackException e) {
            throw new javax.transaction.SystemException("Unexpected RollbackException exception: "+e.getMessage());
        } catch (java.sql.SQLException e) {
            throw new javax.transaction.SystemException("Unexpected SQL exception: "+e.getMessage());
        }
    }

    /**
     *  synchronization implementation
     */
    public void beforeCompletion() {
        // nothing to do
    }

    /**
     *  synchronization implementation
     */
    public void afterCompletion(int status) {
        if (tx == null ) {
            logger.log(BasicLevel.ERROR, "NO TX!");
        }
        pool.freeConnections(tx != null ? tx : enlistedInTx);
    }

    /**
     * @return true if connection max age has expired
     */
    public boolean isAged() {
        return (deathTime < System.currentTimeMillis());
    }

    /**
     * @return true if connection is still open
     */
    public boolean isOpen() {
        return (open > 0);
    }

    /**
     * @return open count
     */
    public int getOpenCount() {
        return open;
    }

    /**
     * Check if the connection has been unused for too long time.
     * This occurs usually when the caller forgot to call close().
     * @return true if open time has been reached, and not involved in a tx.
     */
    public boolean inactive() {
        return (open > 0 && tx == null && enlistedInTx == null && closeTime < System.currentTimeMillis());
    }

    /**
     * @return true if connection is closed
     */
    public boolean isClosed() {
        return (open <= 0);
    }

    /**
     * Notify as opened
     */
    public void open() {
        open++;
        closeTime = System.currentTimeMillis() + pool.getMaxOpenTimeMilli();
    }

    /**
     * notify as closed
     * @return true if normal close.
     */
    public boolean close() {
        open--;
        if (open < 0) {
            logger.log(BasicLevel.WARN, "connection was already closed");
            open = 0;
            return false;
        }
        if (tx == null && open > 0) {
            logger.log(BasicLevel.ERROR, "connection-open counter overflow");
            open = 0;
        }
        return true;
    }

    /**
     * @return the associated XAConnection
     */
    public XAConnection getXACon() {
        return xaConn;
    }

    /**
     * Set the associated transaction
     * @param tx Transaction
     */
    public void setTx(Transaction tx) {
        this.tx = tx;
    }

    /**
     * @return the Transaction
     */
    public Transaction getTx() {
        return tx;
    }

    /**
     * @return true if registered as RME
     */
    public boolean isRME() {
        return rme;
    }

    /**
     * set/unset as RME
     */
    public void setRME(boolean rme) {
        this.rme = rme;
    }

    /**
     * remove this item
     */
    public void remove() {
        // Close the physical connection
        try {
            xaConn.close();
        } catch (java.sql.SQLException ign) {
            logger.log(BasicLevel.ERROR, "Could not close Connection: " + ign);
        }

        // remove all references (for GC)
        xaConn = null;
        tx = null;
        enlistedInTx = null;
    }
}
