/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: PoolKeeper.java,v 1.11 2004/11/17 20:54:07 ehardesty Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.dbm;
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Keeps pool as small as possible (after conn's has reached MaxAge)
 * @author Philippe Durieux
 * Contributor(s):
 */
public class PoolKeeper extends Thread {

    /**
     * Pool object to monitor
     */
    private Pool pool;
    /**
     * Logger object to log events
     */
    private Logger logger = null;
    /**
     * Default adjustment period of 15 seconds
     */
    private long adjustperiod = 15000L;  // default = 15s
    /**
     * Defalut sampling period of 60 seconds
     */
    private long samplingperiod = 60000L;  // default = 60s

    /**
     * Constructor
     * @param pool Pool to monitor
     */
    public PoolKeeper(Pool pool) {
        super("PoolKeeper");
        setDaemon(true);
        this.pool = pool;
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
    }

    /**
     * Set the sampling period.
     * This cannot be done in the PoolKeeper constructor, because the
     * sampling period has not been set already in the Pool.
     * Moreover, this can be reconfigured later.
     * @param sec sampling period in sec.
     */
    public void setSamplingPeriod(int sec) {
        logger.log(BasicLevel.DEBUG, " to " + sec);
        samplingperiod = sec * 1000L;
    }

    /**
     * Start the pool keeper thread
     */
    public void run() {
        long timeout;
        long adjusttime = adjustperiod;
        long samplingtime = samplingperiod;
        while (true) {
            timeout = adjusttime;
            if (samplingtime < timeout) {
                timeout = samplingtime;
            }
            try {
                sleep(timeout);
                adjusttime -= timeout;
                samplingtime -= timeout;
                if (adjusttime <= 0) {
                    pool.adjust();
                    adjusttime = adjustperiod;
                }
                if (samplingtime <= 0) {
                    pool.sampling();
                    samplingtime = samplingperiod;
                }
            } catch (NullPointerException e) {
                logger.log(BasicLevel.ERROR, "poolkeeper NPE:" + e);
                e.printStackTrace();
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "poolkeeper error:" + e);
                e.printStackTrace();
            }
        }
    }
}
