/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JmxServiceImpl.java,v 1.22 2004/10/28 14:58:33 danesa Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.jmx;

import java.net.URI;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.management.ObjectName;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.Context;
import javax.naming.InitialContext;

import mx4j.log.CommonsLogger;

import org.objectweb.carol.util.configuration.CarolConfiguration;
import org.objectweb.carol.util.configuration.CarolCurrentConfiguration;
import org.objectweb.jonas.service.JonasAlreadyStartedException;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JMX Service implementation.
 * Provides specific doStart() and doStop () methods.

 * @author Adriana Danes.
 *
 */

public class JmxServiceImpl extends AbsJmxServiceImpl {

    /**
     * The JNDI name of the JOnAS RMI connector allowing to access the
     * MBeanServer remotelly. Use of this connector was needed before
     * the definition of the Connectors in JMX (before JMX 1.2)
     */
    private String rmiConnectorName = null;

    /**
     * @return The JNDI name of the JOnAS RMI connector
     */
    public String getRmiConnectorName() {
        return this.rmiConnectorName;
    }

    /**
     * Protocol that can be defined by carol configuration
     */
    private static final String JRMP = "jrmp";
    /**
     * Protocol that can be defined by carol configuration
     */
    private static final String IIOP = "iiop";
    /**
     * Protocol that can be defined by carol configuration
     */
    private static final String JEREMIE = "jeremie";
    /**
     * Protocol that can be defined by carol configuration
     */
    private static final String CMI = "cmi";
    /**
     * Connector servers attached to the MBean server.
     * We may have several connector servers only
     * if several communication protocols are used, as defined by the
     * carol configuration.
     */
    private JMXConnectorServer[] connectorServers = null;

    /**
     * The JMXServiceURLs the connector servers are actually listening on.
     */
    private JMXServiceURL[] connectorServerURLs = null;

    /**
     * Init the logger and then use super method.
     * @param ctx configuration for the init method
     * @throws ServiceException if initialization failed
     */
    public void doInit(Context ctx) throws ServiceException {
        // Set the Mx4Logger that redirects log calls to Jakarta Commons Logging.
        mx4j.log.Log.redirectTo(new CommonsLogger());
        super.doInit(ctx);
    }
    /**
     * Start the Service.
     * Only need to create a RMI connector
     * @exception ServiceException the service could not be started
     */
    public void doStart() throws ServiceException {
        String serverName = getJonasServerName();
        try {
            // Create a RMI Connector for the JMX agent using the JOnAS RMIConnector class
            // --------------------------------------------------------------------------
            RMIConnector rmiConnector = new RMIConnectorImpl(getJmxServer());
            // Register the connector in JNDI
            InitialContext ictx = new InitialContext();
            rmiConnectorName = "RMIConnector_" + serverName;
            ictx.bind(rmiConnectorName, rmiConnector);
            ictx.close();

            // Create one or more connector servers
            // (cf. JSR 160, JMX Remote 1.0)
            // ------------------------------------
            // Determine protocols used by Carol and their configuration
            String s = CarolConfiguration.getProtocols();
            StringTokenizer st = new StringTokenizer(s, ",", false);
            HashMap myCarolConfig = new HashMap();
            while (st.hasMoreElements()) {
                CarolCurrentConfiguration carolConfig = CarolCurrentConfiguration.getCurrent();
                String protocol = st.nextToken().trim();
                //if (!protocol.equals(CMI)) {
                    myCarolConfig.put(protocol, carolConfig.getRMIProperties(protocol));
                //}
            }
            int nbProtocols = myCarolConfig.size();
            // Create a JMXServiceURL and a JMXConnectorServer per protocol
            connectorServerURLs = new JMXServiceURL[nbProtocols];
            connectorServers = new JMXConnectorServer[nbProtocols];
            // index over connectorServerURLs
            int i = 0;
            String serviceURL = null;
            // arguments needed to construct a JMXServiceURL
            String myProtocol = null;
            String myHost = null;
            String myPort = null;
            String myName = null;
            for (Iterator it = myCarolConfig.keySet().iterator(); it.hasNext();) {
                String carolProtocol = (String) it.next();
                Properties props = (Properties) myCarolConfig.get(carolProtocol);
                String sCarolURL = props.getProperty(Context.PROVIDER_URL);
                URI carolURL = new URI(sCarolURL);
                myHost = carolURL.getHost();
                myPort = String.valueOf(carolURL.getPort());
                if (carolProtocol.equals(JRMP)) {
                    // Treat JRMP case
                    myProtocol = "jrmp";
                    myName = "jrmpconnector_" + serverName;
                    serviceURL = "service:jmx:rmi://" + myHost + "/jndi/rmi://"
                        + myHost + ":" + myPort + "/"
                        + myName;
                } else if (carolProtocol.equals(JEREMIE)) {
                    // Treat JEREMIE case
                    myProtocol = "jeremie";
                    myName = "jeremieconnector_" + serverName;
                    // PATCH for JEREMIE ONLY !!!!
                    //int anyNumber = 5;
                    //thePort = thePort + anyNumber;
                    serviceURL = "service:jmx:rmi://" + myHost + "/jndi/jrmi://"
                        + myHost + ":" + myPort + "/"
                        + myName;
                } else if (carolProtocol.equals(IIOP)) {
                    // Treat IIOP case
                    myProtocol = "iiop";
                    myName = "iiopconnector_" + serverName;
                    serviceURL = "service:jmx:iiop://" + myHost + "/jndi/iiop://"
                        + myHost + ":" + myPort + "/"
                        + myName;
                } else if (carolProtocol.equals(CMI)) {
                    // Treat JRMP case
                    myProtocol = "cmi";
                    myName = "cmiconnector_" + serverName;
                    serviceURL = "service:jmx:rmi://" + myHost + "/jndi/cmi://"
                        + myHost + ":" + myPort + "/"
                        + myName;
                }
                JMXServiceURL url = new JMXServiceURL(serviceURL);
                // Cast to Map is required for JDK 1.5
                JMXConnectorServer connectorServer = JMXConnectorServerFactory.newJMXConnectorServer(url, (Map) props, null);
                connectorServers[i] = connectorServer;
                // Create the MBean associated to the connector
                String connectorObjectName = "connector_" + carolProtocol;
                ObjectName connectorServerName = JonasObjectName.jmxConnectorServer(myProtocol, connectorObjectName);
                getJmxServer().registerMBean(connectorServer, connectorServerName);
                // Start the JMXConnectorServer
                try {
                    connectorServer.start();
                    connectorServerURLs[i] = connectorServer.getAddress();
                } catch (IllegalArgumentException e) {
                    throw e;
                }
                i++;
            }
        } catch (javax.naming.NameAlreadyBoundException ne) {
            getLogger().log(BasicLevel.DEBUG, "Cannot start JMX service " + ne);
            throw new JonasAlreadyStartedException();
        } catch (Exception e) {
            getLogger().log(BasicLevel.DEBUG, "Cannot start JMX service " + e);
            throw new ServiceException("Cannot start JMX service", e);
        }
    }
    /**
     * Stop this service
     */
    public void doStop() {
        try {
            InitialContext ictx = new InitialContext();
            ictx.unbind("RMIConnector_" + getJonasServerName());
            ictx.close();

            // Deactivates the connector server, that is, stops listening for client connections.
            // Calling this method will also close all client connections that were made by this server.
            // Being a potentialy slow operation, we keep it commented for the moment
            for (int i = 0; i < connectorServers.length; i++) {
                connectorServers[i].stop();
            }
        } catch (Exception e) {
            getLogger().log(BasicLevel.ERROR, "Cannot Unbind Jmx RMI Connector" + e);
        }
        // Unregister some MBeans
        ObjectName domainOn = J2eeObjectName.J2EEDomain(getDomainName());
        try {
            getJmxServer().unregisterMBean(domainOn);
        } catch (Exception e) {
            getLogger().log(BasicLevel.INFO, "Cannot unregister JEEDomain MBean:" + domainOn.toString());
        }
        ObjectName serverOn = J2eeObjectName.J2EEServer(getDomainName(), getJonasServerName());
        try {
            getJmxServer().unregisterMBean(serverOn);
        } catch (Exception e) {
            getLogger().log(BasicLevel.INFO, "Cannot unregister JEEServer MBean:" + serverOn.toString());
        }
        // Remove internal references to the MBeanServer.
        releaseJmxServer();

        getLogger().log(BasicLevel.DEBUG, "JMX Service stopped");
    }


    /**
     *
     * @return The actual adresses on which listen the created connector servers
     */
    public JMXServiceURL[] getConnectorServerURLs() {
        return this.connectorServerURLs;
    }
}
