/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: J2EEDomain.java,v 1.7 2004/11/22 15:02:44 danesa Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.management.j2eemanagement;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.management.JMException;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.J2eeObjectName;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implements a managed object which represents a management domain.
 * @author Adriana Danes
 */
public class J2EEDomain extends J2EEManagedObject {

    /**
     * management domain name
     */
    private String name = null;
    /**
     * List of object names corresponding to the servers associated with this domain
     */
    private ArrayList myServers = null;

    /**
     * J2EE Management logger
     */
    private static Logger logger = null;

    private Map managedServersToUrls = null;
    private Map managedServersToConnectors = null;
    private Map managedServersToConnections = null;

    /**
     * MBean constructor
     * @param objectName object name of the managed object
     * @param stateManageable if true, this managed object implements J2EE State Management Model
     * @param statisticsProvider if true, this managed object implements the J2EE StatisticProvide Model
     * @param eventProvider  if true, this managed object implements the J2EE EventProvider Model
     * @throws JMException ..
     */
    public J2EEDomain(String objectName, boolean stateManageable, boolean statisticsProvider, boolean eventProvider) throws JMException {
        super(objectName, stateManageable, statisticsProvider, eventProvider);
        ObjectName myObjectName = ObjectName.getInstance(objectName);
        name = myObjectName.getKeyProperty("name");
        myServers = new ArrayList();
        managedServersToConnectors = new HashMap();
        managedServersToConnections = new HashMap();
        managedServersToUrls = new HashMap();
        logger = Log.getLogger("org.objectweb.jonas.management.j2eemanagement");
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "J2EEDomain managed object created. OBJECT_NAME = " + objectName);
        }
    }

    /**
     * @return The list of object names corresponding to the servers associated with this domain
     */
    public String[] getServers() {
        //return (String[]) myServers.toArray();
        String[] sb = new String[myServers.size()];
        int i = 0;
        for (Iterator it = myServers.iterator(); it.hasNext();) {
            sb[i++] = (String) it.next();
        }
        return sb;
    }

    /**
     * Add an object name to the <code>servers</code> list.
     * @param serverName ObjectName (String form) of a J2EEServer to be add in this domain
     */
    public void addServer(String serverName) {
        if (myServers.contains(serverName)) {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "The object name: " + serverName + " is already in the servers list");
            }
        } else {
            myServers.add(serverName);
        }
    }

    /**
     * Remove an object name from the <code>servers</code> list.
     * @param serverName ObjectName (String form) of a J2EEServer to be removed from this domain
     * @return Object name of the removed J2EEServer
     */
    public String removeServer(String serverName) {
        int index = myServers.indexOf(serverName);
        if (index == -1) {
            // name not found
            return null;
        } else {
            return (String) myServers.remove(index);
        }
    }

    /**
     * Add a JOnAS server in the management domain in order
     * update the management domain configuration.
     *
     * @param domainName name of the management domain the server belongs
     * @param serverName server name
     * @param connectorServerURL connector server url
     */
    public void addServer(String domainName, String serverName, String connectorServerURL) {
        System.out.println(">>> IN addServer");
        System.out.println("domainName = " + domainName);
        System.out.println("serverName = " + serverName);
        System.out.println("url = " + connectorServerURL);
        if (domainName.equals(name)) {
            // Server thet belongs to the current management domain
            if (managedServersToUrls.containsKey(serverName)) {
                // This server is already known
                ArrayList urls = (ArrayList) managedServersToUrls.get(serverName);
                if (!urls.contains(connectorServerURL)) {
                    // Add this new connectorServerURL in the connectors list
                    urls.add(connectorServerURL);
                    // try to create a connection for this connectorServerURL
                    createConnection(connectorServerURL, serverName);
                }
            } else {
                // Add new server
                ArrayList urls = new ArrayList();
                urls.add(connectorServerURL);
                managedServersToUrls.put(serverName, urls);
                // try to create a connection
                createConnection(connectorServerURL, serverName);
                // Update the servers list
                ObjectName on = J2eeObjectName.J2EEServer(domainName, serverName);
                addServer(on.toString());
            }
        } else {
            if (logger.isLoggable(BasicLevel.WARN)) {
                logger.log(BasicLevel.WARN, "The server named " + serverName + " was not started in the management domain "
                        + name + ", but in management domain " + domainName);
            }
        }
    }

    /**
     * Add a JOnAS server in the management domain
     * @param domainName name of the management domain the server belongs
     * @param serverName server name
     */
    public void removeServer(String domainName, String serverName) {
        if (domainName.equals(name)) {
            // Server thet belongs to the current management domain
            managedServersToConnections.remove(serverName);
            JMXConnector connector = (JMXConnector) managedServersToConnectors.remove(serverName);
            try {
                connector.close();
            } catch (IOException e1) {
            }
            managedServersToUrls.remove(serverName);
        } else {
            if (logger.isLoggable(BasicLevel.WARN)) {
                logger.log(BasicLevel.WARN, "The server named " + serverName + " was not started in the management domain "
                        + name + ", but in management domain " + domainName);
            }
        }
    }
    /**
     * Create a MBeanServer connection using the provided url
     * @param serverName the name of the server
     * @param connectorServerURL the url used to construct connection
     * @return true if a connection created
     */
    private boolean createConnection(String serverName, String connectorServerURL) {
        boolean created = false;
        MBeanServerConnection connection;
        JMXConnector connector = null;
        try {
            JMXServiceURL url = new JMXServiceURL(connectorServerURL);
            connector = JMXConnectorFactory.newJMXConnector(url, null);
            connector.connect(null);
            connection = connector.getMBeanServerConnection();
        } catch (MalformedURLException e) {
            connection = null;
        } catch (IOException e) {
            connection = null;
        }
        if (connection != null) {
            // OK, got a connection for this connector
            if (!managedServersToConnections.containsKey(serverName)) {
                // add new connection and connector
                managedServersToConnections.put(serverName, connection);
                managedServersToConnectors.put(serverName, connector);
                created = true;
                logger.log(BasicLevel.INFO, "First MBeanServerConnection created for connecting to server " + serverName);
            } else {
                // A connection exists already.
                // We could do nothing or create a new one and reolace the old, or manage
                // sever connections.
                // Currently we manage one connection, so we create a new one and replacec the old
                // one with the new one
                // Also, we replace the old connector by the new one
                JMXConnector oldConnector = (JMXConnector) managedServersToConnectors.get(serverName);
                try {
                    oldConnector.close();
                } catch (IOException e1) {
                }
                managedServersToConnectors.put(serverName, connector);
                managedServersToConnections.put(serverName, connection);
                created = true;
                logger.log(BasicLevel.INFO, "New MBeanServerConnection created for connecting to server " + serverName);

            }
        } else {
            if (connector != null) {
                try {
                    connector.close();
                } catch (IOException e1) {
                }
            }
        }
        return created;
    }
}
