/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id: J2EEServerMBean.java,v 1.27 2004/11/23 09:29:29 danesa Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.server;

// Java imports
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Arrays;

import javax.enterprise.deploy.shared.ModuleType;
import javax.management.MBeanException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;

import org.apache.commons.modeler.BaseModelMBean;

import org.objectweb.common.Cmd;

import org.objectweb.jonas_ear.deployment.api.EarDeploymentDesc;
import org.objectweb.jonas_ear.deployment.api.EarDeploymentDescException;
import org.objectweb.jonas_ear.deployment.lib.wrapper.EarManagerWrapper;

import org.objectweb.jonas_ejb.genic.wrapper.GenicServiceWrapper;

import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDescException;
import org.objectweb.jonas_web.deployment.lib.wrapper.WebManagerWrapper;

import org.objectweb.jonas_ws.wsgen.wrapper.WsGenWrapper;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.container.EJBService;
import org.objectweb.jonas.ear.EarService;
import org.objectweb.jonas.ear.EarServiceException;
import org.objectweb.jonas.resource.ResourceService;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.JWebContainerService;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * MBean class for Server management. Two classes are used to provide all
 * management services : J2EEServer and J2EEServerMBean. This class provide the
 * notification's process when add or remove resources.
 * @author Adriana Danes
 * @author Michel-Ange Anton
 */

public class J2EEServerMBean extends BaseModelMBean {

    /**
     * Size of the buffer
     */
    private static final int BUFFER_SIZE = 1024;

    /**
     * Logger
     */
    private static Logger mgtLogger = Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX);

    /**
     * Default constructor
     * @throws MBeanException if super constructor fails
     */
    public J2EEServerMBean() throws MBeanException {
        super();
    }

    // ------------------------------------------------------------- Public
    // methods

    /**
     * All The MBean names corresponding to the deployed J2EEModules (ear, jar,
     * war, rar).
     * @return The String's array of deployed modules objects names
     */
    public String[] getDeployedObjects() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getDeployedObjects();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * All The MBean names corresponding to the deployed J2EEResources.
     * @return The String's array of deployed resources objects names
     */
    public String[] getResources() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getResources();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * All The MBean names corresponding to the used JVMs.
     * @return The String's array of used JVMs objects names
     */
    public String[] getJavaVMs() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getJavaVMs();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * Send a notification to the listener.
     * @param pNotification The notification to send
     */
    public void sendNotification(Notification pNotification) {
        ((J2EEServer) (this.resource)).sendNotification(pNotification);
    }

    /**
     * Add a new listener.
     * @param pListner Listener to notify
     * @param pFilter Notification filter
     * @param pHandback ??
     * @throws java.lang.IllegalArgumentException if notification is not done
     */
    public void addNotificationListener(NotificationListener pListner, NotificationFilter pFilter,
            java.lang.Object pHandback) throws java.lang.IllegalArgumentException {
        ((J2EEServer) (this.resource)).addNotificationListener(pListner, pFilter, pHandback);
    }

    /**
     * Deploy a stand-alone J2EE module packaged in a JAR file
     * @param fileName file name
     * @return The ObjectName of the MBean associated to the deployed module
     * @throws Exception Management operation failed
     */
    public String deployJar(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).deployJar(fileName);
    }

    /**
     * Deploy a stand-alone J2EE module packaged in a War file
     * @param fileName file name
     * @throws Exception Management operation failed
     */
    public void deployWar(String fileName) throws Exception {
        ((J2EEServer) (this.resource)).deployWar(fileName);
    }

    /**
     * Deploy a J2EE application packaged in a EAR file
     * @param fileName file name
     * @return The ObjectName of the MBean associated to the deployed J2EE
     *         Application
     * @throws Exception Management operation failed
     */
    public String deployEar(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).deployEar(fileName);
    }

    /**
     * Deploy a J2EE application packaged in a RAR file
     * @param fileName file name
     * @return The ObjectName of the MBean associated to the deployed J2EE
     *         Application
     * @throws Exception Management operation failed
     */
    public String deployRar(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).deployRar(fileName);
    }

    /**
     * Test if the specified filename is already deployed or not.
     * @param fileName the name of the ear file.
     * @return true if the ear is deployed, otherwise false.
     * @throws Exception Management operation failed
     */
    public Boolean isEarDeployed(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).isEarDeployed(fileName);
    }

    /**
     * Test if the specified filename is already deployed or not.
     * @param fileName the name of the rar file.
     * @return true if the rar is deployed, otherwise false.
     * @throws Exception Management operation failed
     */
    public Boolean isRarDeployed(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).isRarDeployed(fileName);
    }

    /**
     * Test if the specified filename is already deployed or not.
     * @param fileName the name of the jar file.
     * @return true if the jar is deployed, otherwise false.
     * @throws Exception Management operation failed
     */
    public Boolean isJarDeployed(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).isJarDeployed(fileName);
    }

    /**
     * Test if the specified filename is already deployed or not.
     * @param fileName the name of the War file.
     * @return true if the jar is deployed, otherwise false.
     * @throws Exception Management operation failed
     */
    public Boolean isWarDeployed(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).isWarDeployed(fileName);
    }

    /**
     * Deploy file (GenIC), needed for Ishmael to work. The file is in the
     * directory specified by the property "jonas.service.deployment.directory"
     * @param typeparam type of the file (EJB, WAR, EAR, RAR, CAR)
     * @param bfile bytes array of the file
     * @param filename basename of the file to be deployed
     * @param genicArgs arguments for GenIC
     * @throws RemoteException
     * @throws EarServiceException
     * @throws ResourceServiceException
     * @throws JWebContainerServiceException
     * @return absolute path of the file
     * @author Dean Jennings. <da_jennings@junta.com.au>
     */
    public String deployFile(Integer typeparam, java.lang.Byte[] bfile, String filename, String[] genicArgs) {
        String directory = "";
        int type = typeparam.intValue();

        ServiceManager serviceManager = null;
        try {
            serviceManager = ServiceManager.getInstance();
        } catch (Exception e) {
            // TODO
            mgtLogger.log(BasicLevel.ERROR, "Cannot get the instance of the Service Manager : '" + e.getMessage()
                    + "'.");
        }

        if (type == ModuleType.EJB.getValue()) {
                directory = ((EJBService) serviceManager.getEjbService()).getEjbjarsDirectory();
        } else if (type == ModuleType.EAR.getValue()) {
                directory = ((EarService) serviceManager.getEarService()).getAppsDirectory();
        } else if (type == ModuleType.WAR.getValue()) {
                directory = ((JWebContainerService) serviceManager.getWebContainerService()).getWebappsDirectory();
        } else if (type == ModuleType.RAR.getValue()) {
                directory = ((ResourceService) serviceManager.getResourceService()).getRarsDirectory();
        } else if (type == ModuleType.CAR.getValue()) {
            // do nothing
        }

        File file = new File(directory + filename);
        try {
            FileOutputStream out = new FileOutputStream(file);
            byte[] bfileused = new byte[bfile.length];
            for (int i = 0; i < bfile.length; i++) {
                bfileused[i] = bfile[i].byteValue();
            }
            out.write(bfileused);
            out.close();
            if (type == ModuleType.EJB.getValue()) {
                    try {
                        callGenic(genicArgs, directory + filename);
                    } catch (ServiceException e) {
                        mgtLogger.log(BasicLevel.WARN, "deploy File error " + filename + e.getClass().getName() + " "
                                + e.getMessage());
                    }
            } else if (type == ModuleType.EAR.getValue()) {
                unpackAndCompileEar(file, genicArgs);
            } else if (type == ModuleType.WAR.getValue()) {
                 checkWebAppDeploymentDesc(file);
            } else {
                mgtLogger.log(BasicLevel.DEBUG, "Deployment not yet implemented for type '" + type + ".");
            }
        } catch (IOException ioe) {
            mgtLogger.log(BasicLevel.ERROR, "Cannot dump to outputstream the file '" + file + "' : '"
                    + ioe.getMessage() + "'.");
            // TODO
        }

        if (!filename.endsWith(".rar")) {
            String modifiedArchive = null;
            WsGenWrapper ww = new WsGenWrapper();
            try {
                modifiedArchive = ww.callWsGenExecute(file.getPath());
            } catch (Exception e) {
                mgtLogger.log(BasicLevel.ERROR, "Cannot launch WsGen on the file '" + file + "' : '"
                        + e.getCause().getMessage() + "'.");
            }
            boolean modified = false;
            try {
                modified = ww.callWsGenIsInputModifed();
            } catch (Exception e) {
                mgtLogger.log(BasicLevel.ERROR, "Cannot launch WsGen.isInputModified '" + file + "' : '"
                        + e.getCause().getMessage() + "'.");
                // assume file has been modified
                modified = true;
            }
            if (modified) {
                File newFile = new File(modifiedArchive);

                mgtLogger.log(BasicLevel.DEBUG, "Copying file " + newFile + " to file " + file);
                try {
                    copy(newFile, file);
                } catch (Exception e) {
                    mgtLogger.log(BasicLevel.ERROR, "Error when copying file '" + newFile + "'. : " + e.getMessage());
                }
                return modifiedArchive;
            } else {
                return file.getPath();
            }
        } else {
            return file.getPath();
        }
    }

    /**
     * Copy src file to dst file.
     * @param src file to copy
     * @param dst file where the file need to be copied
     * @throws IOException if the copy fails
     */
    void copy(File src, File dst) throws IOException {
        InputStream is = new FileInputStream(src);
        OutputStream os = new FileOutputStream(dst);

        // bytes to transfer
        byte[] buf = new byte[BUFFER_SIZE];
        int len;
        while ((len = is.read(buf)) > 0) {
            os.write(buf, 0, len);
        }
        is.close();
        os.close();
    }

    /**
     * Call GenIC by using a wrapper
     * @param genicArgs arguments for GenIC
     * @param jarPath path of the jar file
     */
    private void callGenic(String[] genicArgs, String jarPath) {
        String[] args;
        if (genicArgs != null) {
            args = new String[genicArgs.length + 1];
            for (int i = 0; i < genicArgs.length; i++) {
                args[i] = genicArgs[i];
            }
            args[genicArgs.length] = jarPath;
        } else {
            args = new String[1];
            args[0] = jarPath;
        }
        mgtLogger.log(BasicLevel.DEBUG, "Calling GenIC with arguments :" + Arrays.asList(args));
        GenicServiceWrapper.callGenic(args);
    }

    /**
     * Unpack and then launch GenIC on each ejb jar files
     * @param file ear file
     * @param genicArgs arguments for GenIC
     * @throws ServiceException if this step fails
     */
    private void unpackAndCompileEar(File file, String[] genicArgs) throws ServiceException {
        URL[] earUrl = new URL[1];

        try {
            earUrl[0] = file.toURL();
        } catch (MalformedURLException e) {
            // TODO
            mgtLogger.log(BasicLevel.ERROR, "The url for the file '" + file + "' cannot be built : '" + e.getMessage()
                    + "'.");
        }

        //Create classLoader
        //parent classloader is the current classloader
        ClassLoader currentLoader = Thread.currentThread().getContextClassLoader();
        URLClassLoader loaderCls = new URLClassLoader(earUrl, currentLoader);

        EarDeploymentDesc desc = null;
        try {
            desc = EarManagerWrapper.getDeploymentDesc(earUrl[0].getFile(), loaderCls);
        } catch (EarDeploymentDescException e) {
            // TODO
            mgtLogger.log(BasicLevel.ERROR, "Cannot get deployment descriptor for the Ear file '" + file + "' : '"
                    + e.getMessage() + "'.");
        }

        String[] ejbTags = desc.getEjbTags();
        if (ejbTags.length != 0) {
            //  need to call GENIC
            /*
             * Example of a jar command: jar -uf ../output/ejbjars/sb.jar -C
             * /tmp/genic1547.tmp a/b/C1.class -C /tmp/genic1547.tmp
             * a/b/C2.class .....
             */
            String javaHomeBin = System.getProperty("java.home", "");
            if (!("".equals(javaHomeBin))) {
                javaHomeBin = javaHomeBin + File.separator + ".." + File.separator + "bin" + File.separator;
            }

            Cmd cmd = null;
            cmd = new Cmd(javaHomeBin + "jar");
            cmd.addArgument("-xf");
            cmd.addArgument(file.getAbsolutePath());

            for (int i = 0; i < ejbTags.length; i++) {
                cmd.addArgument(ejbTags[i]);
            }

            boolean exitCmd = cmd.run();
            // Analyse the result of the jar command
            if (!exitCmd) {
                throw new EarServiceException("Failed when extracting the the ejb jar " + "in the given jar file '"
                        + file + "'.");
            }

            for (int i = 0; i < ejbTags.length; i++) {
                try {
                    callGenic(genicArgs, ejbTags[i]);
                } catch (ServiceException e) {
                    throw (e);
                }
            }
            // add modified jar into ear
            cmd = new Cmd(javaHomeBin + "jar");
            cmd.addArgument("-uf");
            cmd.addArgument(file.getAbsolutePath());

            for (int i = 0; i < ejbTags.length; i++) {
                cmd.addArgument(ejbTags[i]);
            }
            exitCmd = cmd.run();
            // Analyse the result of the jar command
            if (!exitCmd) {
                throw new EarServiceException("Failed when extracting the the ejb jar " + "in the given jar file '"
                        + file + "'.");
            }

            for (int i = 0; i < ejbTags.length; i++) {
                File del = new File(ejbTags[i]);
                del.delete();
            }
        }
    }

    /**
     * Validate deployment descriptor of the War file
     * @param file given file
     */
    private void checkWebAppDeploymentDesc(File file) {
        URL warUrl = null;

        try {
            warUrl = file.toURL();
        } catch (MalformedURLException mue) {
            mgtLogger.log(BasicLevel.ERROR, "The url for the file '" + file + "' cannot be built : '"
                    + mue.getMessage() + "'.");
            throw new RuntimeException(mue);
        }

        URLClassLoader warCl = new URLClassLoader(new URL[] {warUrl});

        try {
            WebManagerWrapper.getDeploymentDesc(warUrl, warCl, null);
        } catch (WebContainerDeploymentDescException e) {
            String err = "Cannot read the deployment descriptors '" + warUrl.getFile() + "'";
            mgtLogger.log(BasicLevel.ERROR, err + ": " + e);
            e.printStackTrace(System.err);
            throw new RuntimeException(err, e);
        }

    }

    /**
     * Add an object name to the <code>resources</code> list.
     * @param pObjectName Object name correspondig to a J2EEResource MBean
     */
    //
    //    public static void addResource(String pObjectName) {
    //        try {
    //            Server oServer = Server.getInstance();
    //            J2EEServer oJ2EEServer = oServer.getJ2EEServer();
    //            oJ2EEServer.addResource(pObjectName);
    //        } catch (Exception e) {
    //            mgtLogger.log(BasicLevel.WARN
    //                , "Add resource error (object=" + pObjectName + ") : "
    //                + e.getClass().getName() + " " + e.getMessage());
    //        }
    //    }
    /**
     * Remove an object name from the <code>resources</code> list.
     * @param pObjectName Object name correspondig to a J2EEResource MBean
     * @return Object name to the removed J2EEResource MBean
     */

    //    public static String removeResource(String pObjectName) {
    //        String sRet = null;
    //        try {
    //            Server oServer = Server.getInstance();
    //            J2EEServer oJ2EEServer = oServer.getJ2EEServer();
    //            sRet = oJ2EEServer.removeResource(pObjectName);
    //        } catch (Exception e) {
    //            mgtLogger.log(BasicLevel.WARN
    //                , "Remove resource object error (object=" + pObjectName + ") : "
    //                + e.getClass().getName() + " " + e.getMessage());
    //        }
    //        return sRet;
    //    }
}