/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Takoua Abdellatif.
 * Contributor(s): _______________________.
 */

package org.objectweb.jonas.server.discovery;

import java.net.InetAddress;
import java.net.UnknownHostException;

import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.NotificationBroadcasterSupport;
import javax.management.ObjectName;

import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * The DiscoveryClient is in charge of sending discovery messages on the LAN to
 * discover resources. It waits during a Timeout period of time for the
 * discovery answers.
 *
 * @author <a href="mailto:Takoua.Abdellatif@inria.fr">Takoua Abdellatif </a>
 * @version 1.0
 */
public class DiscoveryClient extends NotificationBroadcasterSupport implements
	DiscoveryClientMBean,
	MBeanRegistration {

    private int                     sourcePort              = 9888;
    private String                  sourceIp                = null;
    private MBeanServer             mbeanServer;
    private ObjectName              myOn;
    private Thread                  discoveryClientListener = null;
    private DiscoveryClientListener dcl                     = null;
    private int                     ttl                     = 1;
    private int                     timeout                 = 1000;
    private int                     listeningPort           = 9080;
    private String                  listeningIp             = "224.224.224.224";
    private static Logger           logger                  = Log
    .getLogger(Log.JONAS_SERVER_PREFIX);

    public DiscoveryClient() {
        try {
            JProp props = JProp.getInstance();
            String port = props.getValue("discoveryClientPort");
            if (port != null) {
                this.sourcePort = (Integer.valueOf(port)).intValue();
            }
            System.out.println("Source port is: " + sourcePort);
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     *
     * @see org.objectweb.jonas.server.discovery.DiscoveryClientMBean#getTimeOut()
     */
    public int getTimeout() {
        return timeout;
    }

    /**
     *
     * @see org.objectweb.jonas.server.discovery.DiscoveryClientMBean#setTimeout(int)
     */
    public void setTimeout(int timeout) {
        this.timeout = timeout;
    }

    /**
     *
     * @see org.objectweb.jonas.server.discovery.EnrollerMBean#getListeningPort()
     */
    public int getListeningPort() {
        return listeningPort;
    }

    /**
     *
     * @see org.objectweb.jonas.server.discovery.EnrollerMBean#setListeningPort(int)
     */
    public void setListeningPort(int listeningPort) {
        this.listeningPort = listeningPort;

    }

    /**
     * @see org.objectweb.jonas.server.discovery.EnrollerMBean#getListeningIp()
     */
    public String getListeningIp() {
        return listeningIp;
    }

    /**
     *
     * @see org.objectweb.jonas.server.discovery.EnrollerMBean#setListeningIP(java.lang.String)
     */
    public void setListeningIP(String IpAddress) {
        this.listeningIp = IpAddress;

    }

    /**
     * @see org.objectweb.jonas.server.discovery.DiscoveryClientMBean#getSourcePort()
     */
    public int getSourcePort() {
        return sourcePort;
    }

    /**
     * @see org.objectweb.jonas.server.discovery.DiscoveryClientMBean#setSourcePort(int)
     */
    public void setSourcePort(int sourcePort) {
        this.sourcePort = sourcePort;

    }

    /**
     * @see org.objectweb.jonas.server.discovery.DiscoveryClientMBean#getSourceIp()
     */
    public String getSourceIp(){
        return sourceIp;
    }

    /**
     * @see org.objectweb.jonas.server.discovery.EnrollerMBean#getTimeToLive()
     */
    public int getTimeToLive() {
        return ttl;
    }

    /**
     * @see org.objectweb.jonas.server.discovery.DiscoveryClientMBean#setSourceIp(java.lang.String)
     */
    public void setSourceIp(String sourceIp) {
        this.sourceIp = sourceIp;

    }

    /**
     *
     * @see org.objectweb.jonas.server.discovery.EnrollerMBean#setTimeToLive(int)
     */
    public void setTimeToLive(int ttl) {
        this.ttl = ttl;

    }

    /**
     *
     * @see org.objectweb.jonas.server.discovery.EnrollerMBean#start()
     */
    public void start() {
        // create the thread in charge of sending the discovery message and waiting
        // for the responses
        // creates an instance of the Discovery communication class
        try {
            this.setSourceIp(InetAddress.getLocalHost().getHostAddress());
        } catch (UnknownHostException e) {
            logger.log(BasicLevel.ERROR, "Unable to create a localhost.");
            e.printStackTrace();
        }
        dcl = new DiscoveryClientListener(this);

        if (discoveryClientListener == null) {
            discoveryClientListener = new Thread(dcl, "discoveryClientListener");
        }
        discoveryClientListener.start();

    }

    /**
     *
     * @see javax.management.MBeanRegistration#preRegister(javax.management.MBeanServer,
     *      javax.management.ObjectName)
     */
    public ObjectName preRegister(MBeanServer mbeanServer, ObjectName on) {
        //  set mbeanServer
        this.mbeanServer = mbeanServer;
        this.myOn = on;

        return myOn;
    }

    /**
     *
     * @see javax.management.MBeanRegistration#postRegister(java.lang.Boolean)
     */
    public void postRegister(Boolean arg0) {

    }

    /**
     *
     * @see javax.management.MBeanRegistration#preDeregister()
     */
    public void preDeregister() throws Exception {
        dcl.stop();

    }

    /**
     *
     * @see javax.management.MBeanRegistration#postDeregister()
     */
    public void postDeregister() {
    }

}