/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Takoua Abdellatif.
 * Contributor(s): _______________________.
 */

package org.objectweb.jonas.server.discovery;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.ServerSocket;
import java.net.SocketException;
import java.net.UnknownHostException;
import javax.management.Notification;
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * @author <a href="mailto:Takoua.Abdellatif@inria.fr">Takoua Abdellatif </a>
 * @version 1.0
 */
public class DiscoveryClientListener implements Runnable {
    private static int      RECEIVE_BUFFER_SIZE = 1024;
    private int             port;
    private InetAddress     destAddress;
    private String          sourceIp;
    private int             sourcePort;
    private DiscoveryClient discoveryClient;
    private int             ttl;
    private MulticastSocket multicastSocket;
    private DatagramSocket  unicastSocket;
    private boolean         notStopped          = true;
    private long            timeout             = 1000;
    private long            sequenceNumber      = 0;
    private static Logger   logger              = Log
    .getLogger(Log.JONAS_SERVER_PREFIX);

    public DiscoveryClientListener(DiscoveryClient discoveryClient) {
        this.port = discoveryClient.getListeningPort();
        this.discoveryClient = discoveryClient;
        this.timeout = discoveryClient.getTimeout();
        this.sourcePort = discoveryClient.getSourcePort();
        this.sourceIp = discoveryClient.getSourceIp();
        try {
            this.destAddress = InetAddress.getByName(discoveryClient.getListeningIp());
            this.ttl = discoveryClient.getTimeToLive();
        } catch (UnknownHostException e) {
            e.printStackTrace();
        }

    }

    /**
     * sends a notification message to the server group.
     */
    public void sendDiscoveryMessage(DiscMessage msg) {
        String theHostName;
        try {
            //send it on the multicast address
            //after transforming the object to a datagram
            logger.log(BasicLevel.DEBUG, "DiscoveryClient : The message to send is"
                    + msg);
            //multicastSocket.send(DiscoveryMessage.getDatagram(new String("toto")));
            byte[] messageBytes = DiscMessage.objectToBytes(msg);
            multicastSocket.send(new DatagramPacket(messageBytes,
                    messageBytes.length, destAddress, port));
        } catch (IOException e1) {
            logger.log(BasicLevel.ERROR,
            "DiscoveryClient :  Error to send an event to Discovery service");
            e1.printStackTrace();
        }
    }

    /**
     * @see java.lang.Runnable#run()
     */
    public void run() {

        Object objReceived = null;
        ObjectInputStream in = null;
        ServerSocket sock = null;
        DatagramPacket datagram = new DatagramPacket(new byte[RECEIVE_BUFFER_SIZE],
                RECEIVE_BUFFER_SIZE);

        // Create a multicast socket
        try {
            multicastSocket = new MulticastSocket(port);
            multicastSocket.setTimeToLive(ttl);
            multicastSocket.joinGroup(destAddress);
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        // Create a unicast socket to receive responses

        try {
            unicastSocket = new DatagramSocket(sourcePort);

        } catch (SocketException e2) {
            logger.log(BasicLevel.ERROR,
            "DiscoveryClient : Unable to create a Datagram socket");
            e2.printStackTrace();
            return;
        }

        // Prepare a discovery message
        DiscMessage msg = new DiscMessage(sourceIp, sourcePort);
        //  First send a discovery message
        sendDiscoveryMessage(msg);
        logger.log(BasicLevel.DEBUG, " DiscoveryClient: Message to send is" + msg);

        //  wait for responses during a Timeout period on the unicast socket.
        long lastTime = timeout + System.currentTimeMillis();
        DiscEvent event;
        try {
            while ((notStopped) && System.currentTimeMillis() <= lastTime) {
                objReceived = null;
                unicastSocket.receive(datagram);

                in = new ObjectInputStream(new ByteArrayInputStream(datagram.getData()));

                objReceived = in.readObject();
                logger.log(BasicLevel.DEBUG, "DiscoveryClient : a message is received"
                        + objReceived);
                if (objReceived != null) {
                    if (objReceived instanceof DiscEvent) {
                        event = (DiscEvent) objReceived;
                        handleReceivedMessage(event);
                    }
                }
            }
        } catch (SocketException e) {
            logger.log(BasicLevel.ERROR, "DiscoveryClient : Socket closed" + e);
            notStopped = false;
        } catch (IOException e1) {
            e1.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
    }

    /**
     * @param msg
     *          a discovery message response received.
     */
    private void handleReceivedMessage(DiscEvent msg) {
        logger.log(BasicLevel.DEBUG,
                "DiscoveryClient : Received a notification message" + msg);
        // create a notification message for all listeners of the event
        // The notification state is the same as Jonas state in the sent event.
        // The notification is only sent if the event comes from a remote machine.
        if (!msg.getSourceAddress().equals(sourceIp)) {
            Notification notif = new Notification(msg.getState(), discoveryClient,
                    sequenceNumber++, System.currentTimeMillis());
            notif.setUserData(msg);
            discoveryClient.sendNotification(notif);
        }
    }

    /**
     * Stops the current thread
     */
    public void stop() {
        notStopped = false;
        Thread.interrupted();
    }
}