/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Takoua Abdellatif.
 * Contributor(s): _______________________.
 */

package org.objectweb.jonas.server.discovery;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.SocketException;
import java.net.UnknownHostException;

import javax.management.MBeanServer;
import javax.management.remote.JMXServiceURL;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * @author <a href="mailto:Takoua.Abdellatif@inria.fr">Takoua Abdellatif </a>
 * @version 1.0
 */
public class DiscoveryComm implements Runnable {
    private int              RECEIVE_BUFFER_SIZE = 1024;
    private MulticastSocket  multicastSocket;
    private DiscoveryManager dm;
    private int              port;
    private InetAddress      destAddress;
    private boolean          notStopped          = true;
    private int              ttl                 = 1;
    private static long      sequenceNumber      = 0;
    private static Logger    logger              = Log.getLogger(Log.JONAS_SERVER_PREFIX);

    // set to false when the thread is stopped.

    public DiscoveryComm(DiscoveryManager dm) {
        this.port = dm.getListeningPort();
        try {
            this.destAddress = InetAddress.getByName(dm.getListeningIp());
            this.ttl = dm.getTimeToLive();
        } catch (UnknownHostException e) {
            e.printStackTrace();
        }
        this.dm = dm;
    }

    private void join() {
        try {
            multicastSocket = new MulticastSocket(port);
            multicastSocket.setTimeToLive(ttl);
            multicastSocket.joinGroup(destAddress);
            logger.log(BasicLevel.DEBUG, " DiscoveryComm: multicast ip address is"
                    + destAddress);

            logger.log(BasicLevel.DEBUG, " DiscoveryComm: multicast port is" + port);

        } catch (IOException e) {
            logger.log(BasicLevel.ERROR, "DiscoveryComm: io problem");
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * sends a notification message to the server group.
     */
    public void sendNotif(DiscEvent msg) {
        String theHostName;
        //waitServerReady();
        try {
            //send it on the multicast address
            //after transforming the object to a datagram
            logger.log(BasicLevel.DEBUG, "DiscoveryComm : The message to send is "
                    + msg);
            byte[] messageBytes = DiscEvent.objectToBytes(msg);
            multicastSocket.send(new DatagramPacket(messageBytes,
                    messageBytes.length, destAddress, port));
        } catch (IOException e1) {
            logger.log(BasicLevel.ERROR, "DiscoveryComm: Error to send notification");
            e1.printStackTrace();
        }

    }

    /**
     * @param state
     *          takes either RUNNING or STOPPING values in the case of a
     *          notification service, depending on the server state.
     * @return a Discovery notification
     * @throws Exception
     *           is thrown if the jmx service is not reached.
     */
    public DiscEvent createNotifMessage(String state) throws Exception {
        String theHostName;
        MBeanServer mbeanServer;
        try {
            theHostName = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            e.printStackTrace();
            return null;
        }
        int theJndiPort = 1099;
        JmxService jmxService;
        // Get jonas information from the MBean server
        try {
            jmxService = (JmxService) ServiceManager.getInstance().getJmxService();
        } catch (Exception se) {
            logger.log(BasicLevel.ERROR, "Could not access the jmx service");
            throw se;
        }

        String serverName = jmxService.getJonasServerName();
        String domainName = jmxService.getDomainName();
        JMXServiceURL[] urls = null;
        // urls are sent only when Jonas is running.
        if (state.equals(DiscEvent.RUNNING))
            urls = jmxService.getConnectorServerURLs();
        // In the case of a notification, the field port is not important since the
        // notifier is not waiting for an acknowledfement.
        DiscEvent resp = new DiscEvent(theHostName, port, serverName, domainName,
                urls);
        resp.setState(state);
        return resp;
    }

    /**
     *
     * @see java.lang.Runnable#run()
     */
    public void run()  {
        logger.log(BasicLevel.DEBUG,
        "DiscoveryComm: run called in DiscoveryComm class");
        DatagramPacket datagram = new DatagramPacket(new byte[RECEIVE_BUFFER_SIZE],
                RECEIVE_BUFFER_SIZE);
        Object objReceived;
        ObjectInputStream in;
        //
        join();
        DiscEvent msg = null;
        try {
            msg = createNotifMessage(DiscEvent.RUNNING);
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR,
            "DiscoveryComm:  Unable to create a notification message");
            e.printStackTrace();
        } if (msg != null)
            sendNotif(msg);
        DiscMessage request;
        DatagramSocket unicastSocket;
        try {
            unicastSocket = new DatagramSocket();
        } catch (SocketException e3) {
            e3.printStackTrace();
            return;
        }
        ;
        try {
            while (notStopped) {
                objReceived = null;
                multicastSocket.receive(datagram);

                in = new ObjectInputStream(new ByteArrayInputStream(datagram.getData()));
                objReceived = in.readObject();

                if (objReceived != null) {
                    // The Event messages are ignored
                    if (objReceived instanceof DiscMessage) {
                        if (objReceived instanceof DiscEvent)
                            logger.log(BasicLevel.DEBUG,
                                    "DiscoveryComm: This message is ignored" + objReceived);
                        else {
                            request = (DiscMessage) objReceived;
                            logger.log(BasicLevel.DEBUG,
                                    "DiscoveryComm: A dicovery message is received "
                                    + objReceived);

                            byte[] messageBytes = null;
                            try {
                                messageBytes = DiscMessage
                                .objectToBytes(createNotifMessage(DiscEvent.RUNNING));
                            } catch (Exception e2) {
                                // TODO Auto-generated catch block
                                e2.printStackTrace();
                            }
                            if (messageBytes != null) {
                                // send the unicast response to the discovery client
                                unicastSocket.send(new DatagramPacket(messageBytes,
                                        messageBytes.length, InetAddress.getByName(request
                                                .getSourceAddress()), request.getSourcePort()));
                            }

                        }
                    }
                }
            }
        } catch (SocketException e) {
            logger.log(BasicLevel.ERROR,"DiscoveryComm: Socket closed" + e);
            notStopped = false;
        } catch (IOException e1) {
            e1.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
    }

    /**
     * sends a notification message to notify that the server is stopping.
     *
     */
    public void stop() {
        // send a notification message of type STOPPING
        DiscEvent msg = null;
        try {
            msg = createNotifMessage(DiscEvent.STOPPING);
        } catch (Exception e) {
            e.printStackTrace();
        }
        if (msg != null) {
            sendNotif(msg);
        }
        Thread.interrupted();
    }

}