/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JAxisServlet.java,v 1.2 2004/09/29 15:18:10 sauthieg Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.ws.axis;

import java.util.Map;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServlet;

import org.apache.axis.AxisFault;
import org.apache.axis.server.AxisServer;
import org.apache.axis.transport.http.AxisServlet;

/**
 * JOnAS AxisServlet version (add multiple AxisEngine in ServletContext).
 * @author Guillaume Sauthier
 */
public class JAxisServlet extends AxisServlet {

    /**
     * Destroy method is called when the servlet is going away. Pass this down
     * to the AxisEngine to let it clean up... <br/>
     * But don't create the engine if it hasn't already been created.
     * TODO Fixme for multiple servlets. This
     * has always been slightly broken (the context's copy stayed around), but
     * now we have extracted it into a superclass it is blatantly broken.
     */
    public void destroy() {

        //if we have had anything to do with creating an axis server
        if (axisServer != null) {
            //then we lock it
            synchronized (axisServer) {
                //clean it up
                axisServer.cleanup();
                //and erase our history of it
                axisServer = null;
                storeEngine(this, null);
            }
        }
    }

    /**
     * get the engine for this servlet from cache or context
     * @return the current servlet's engine
     * @throws AxisFault when an AxisEngine cannot be found
     */
    public AxisServer getEngine() throws AxisFault {
        if (axisServer == null) {
            axisServer = getEngine(this);
        }
        return axisServer;
    }

    /**
     * This is a uniform method of initializing AxisServer in a servlet context.
     * @return Returns the AxisEngine given the current servlet environnment
     * @param servlet servlet wanting an AxisEngine
     * @throws AxisFault when an AxisEngine cannot be found TODO add catch for
     *         not being able to cast the context attr to an engine and reinit
     *         the engine if so.
     */
    public static AxisServer getEngine(HttpServlet servlet) throws AxisFault {
        AxisServer engine = null;

        synchronized (servlet) {
            engine = retrieveEngine(servlet);
            if (engine == null) {
                Map environment = getEngineEnvironment(servlet);

                // Obtain an AxisServer by using whatever AxisServerFactory is
                // registered. The default one will just use the provider we
                // passed in, and presumably JNDI ones will use the
                // ServletContext
                // to figure out a JNDI name to look up.
                //
                // The point of doing this rather than just creating the server
                // manually with the provider above is that we will then support
                // configurations where the server instance is managed by the
                // container, and pre-registered in JNDI at deployment time. It
                // also means we put the standard configuration pattern in one
                // place.
                engine = AxisServer.getServer(environment);
                // attach the AxisServer with the current Servlet
                engine.setName(servlet.getServletName());
                storeEngine(servlet, engine);
            }
        }

        return engine;
    }

    /**
     * put the engine back in to the context.
     * @param servlet servlet wanting to store an engine
     * @param engine reference to the engine. If null, the engine is removed
     */
    private static void storeEngine(HttpServlet servlet, AxisServer engine) {
        ServletContext context = servlet.getServletContext();
        String axisServletName = servlet.getServletName();
        if (engine == null) {
            context.removeAttribute(axisServletName + ATTR_AXIS_ENGINE);
            // find if there is other AxisEngine in Context
            AxisServer server = (AxisServer) context.getAttribute(ATTR_AXIS_ENGINE);

            // no other AxisEngine in ServletContext
            if (servlet.getServletName().equals(server.getName())) {
                context.removeAttribute(ATTR_AXIS_ENGINE);
            }
        } else {
            if (context.getAttribute(ATTR_AXIS_ENGINE) == null) {
                // first Axis servlet to store its AxisEngine
                // use default name
                context.setAttribute(ATTR_AXIS_ENGINE, engine);
            }
            context.setAttribute(axisServletName + ATTR_AXIS_ENGINE, engine);
        }
    }

    /**
     * Get an engine from the servlet context; robust againt serialization
     * issues of hot-updated webapps. Remember than if a webapp is marked as
     * distributed, there is more than 1 servlet context, hence more than one
     * AxisEngine instance.
     * @param servlet the servlet wanting to store/remove an AxisEngine
     * @return the engine or null if either the engine couldnt be found or the
     *         attribute wasnt of the right type
     */
    private static AxisServer retrieveEngine(HttpServlet servlet) {
        Object contextObject = servlet.getServletContext().getAttribute(servlet.getServletName() + ATTR_AXIS_ENGINE);
        if (contextObject == null) {
            // if AxisServer not found :
            // fall back to the "default" AxisEngine
            contextObject = servlet.getServletContext().getAttribute(ATTR_AXIS_ENGINE);
        }
        if (contextObject instanceof AxisServer) {
            AxisServer server = (AxisServer) contextObject;
            // if this is "our" Engine
            if (servlet.getServletName().equals(server.getName())) {
                return server;
            }
            return null;
        } else {
            return null;
        }
    }

}