/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JOnASEJBProvider.java,v 1.3 2004/09/09 16:12:36 sauthieg Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.ws.axis;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;

import org.apache.axis.AxisFault;
import org.apache.axis.Handler;
import org.apache.axis.MessageContext;
import org.apache.axis.i18n.Messages;
import org.apache.axis.providers.java.RPCProvider;
import org.apache.axis.utils.ClassUtils;

/**
 * Expose the service-endpoint interface of the Ejb.
 *
 * @author Guillaume Sauthier
 */
public class JOnASEJBProvider extends RPCProvider {

    /**
     * parameter service-endpoint class name
     */
    public static final String OPTION_SEINTERFACENAME = "serviceEndpointInterfaceName";

    /**
     * parameter service-endpoint JNDI name
     */
    public static final String OPTION_SEJNDINAME = "serviceEndpointJndiName";

    /**
     * cached initial context
     */
    private static InitialContext cachedContext;

    /**
     * Create a new JOnASEJBProvider
     */
    public JOnASEJBProvider() {
        super();
    }

    /**
     * Return a object which implements the service.
     *
     * @param msgContext the message context
     * @param seiName The Service Endpoint Interface classname
     *
     * @return an object that implements the service
     *
     * @throws Exception when trying to create a Serviceobject without serviceendpoint name aprameter
     */
    protected Object makeNewServiceObject(MessageContext msgContext, String seiName) throws Exception {
        //String serviceEndpointName = getStrOption(OPTION_SEINTERFACENAME, msgContext.getService());

        if (seiName == null) {
            // cannot find service-endpoint
            throw new AxisFault(Messages.getMessage("noOption00", OPTION_SEINTERFACENAME, msgContext.getService()
                    .getName()));
        }

        // Load the Home class name given in the config file
        Class seInterface = ClassUtils.forName(seiName, true, msgContext.getClassLoader());

        return createServiceEndpointEJB(msgContext, seInterface);
    }

    /**
     * Returns the Object where methods can be invoked.
     *
     * @param msgContext MessageContext
     * @param seInterface service endpoint interface
     *
     * @throws Exception when paramerters are missing
     *
     * @return Returns the Object where methods can be invoked.
     */
    private Object createServiceEndpointEJB(MessageContext msgContext, Class seInterface) throws Exception {
        String jndiName = getStrOption(OPTION_SEJNDINAME, msgContext.getService());
        if (jndiName == null) {
            throw new AxisFault("Missing parameter in service : " + OPTION_SEJNDINAME);
        }
        Object o = null;
        try {
            InitialContext ic = getCachedContext();
            o = ic.lookup(jndiName);
            o = PortableRemoteObject.narrow(o, seInterface);
        } catch (NamingException ne) {
            throw new AxisFault("Cannot lookup ServiceEndpoint bean : " + jndiName);
        }

        if (seInterface.isInstance(o)) {
            return o;
        } else {
            throw new AxisFault("Object retrieved from lookup not instance of : " + seInterface);
        }
    }

    /**
     * @return Return the option in the configuration that contains the service class
     * name. In the EJB case, it is the JNDI name of the bean.
     */
    protected String getServiceClassNameOptionName() {
        return OPTION_SEINTERFACENAME;
    }

    /**
     * Get a String option by looking first in the service options, and then at
     * the Handler's options. This allows defaults to be specified at the
     * provider level, and then overriden for particular services.
     *
     * @param optionName the option to retrieve
     * @param service Option holder
     *
     * @return String the value of the option or null if not found in either
     *         scope
     */
    protected String getStrOption(String optionName, Handler service) {
        String value = null;
        if (service != null) {
            value = (String) service.getOption(optionName);
        }
        if (value == null) {
            value = (String) getOption(optionName);
        }
        return value;
    }

    /**
     * @return Returns the cached InitialContext (or created a new one)
     * @throws javax.naming.NamingException when InitialContext creation fails
     */
    protected InitialContext getCachedContext() throws javax.naming.NamingException {
        if (cachedContext == null) {
            cachedContext = new InitialContext();
        }
        return cachedContext;
    }

}