/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntityHandle.java,v 1.7 2004/11/09 16:26:13 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.io.IOException;
import java.io.Serializable;
import java.rmi.RemoteException;

import javax.ejb.EJBObject;
import javax.ejb.Handle;
import javax.ejb.HomeHandle;
import javax.ejb.spi.HandleDelegate;
import javax.naming.NamingException;

import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class implements javax.ejb.Handle interface. For a Entity Bean a Handle
 * is a serializable class that contains the HomeHandle and the primary key
 * @author Philippe Coq
 */
public abstract class JEntityHandle implements Handle, Serializable {

    /**
     * @serial
     */
    protected HomeHandle homehandle = null;

    /**
     * @serial
     */
    protected Serializable pk = null;

    /**
     * HandleDelegate used for Serialization.
     */
    private HandleDelegate hdld = null;

    /**
     * EJBObject is computed in derived class.
     */
    protected EJBObject remote = null;
    
    /**
     * constructor
     * @param remote The Remote object
     */
    public JEntityHandle(JEntityRemote jremote) {
        try {
            homehandle = jremote.getEJBHome().getHomeHandle();
        } catch (RemoteException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "cannot get HomeHandle: ", e);
        }
    }

    /**
     * Obtains the EJB object represented by this handle.
     * @return The EJB object
     * @throws RemoteException The EJB object could not be obtained because of a
     *         system-level failure.
     */
    public abstract EJBObject getEJBObject() throws RemoteException;

    /**
     * @return the Primary Key embedded in the Handle
     */
    public java.lang.Object getPK() {
        return pk;
    }
    
    /**
     * Specific implementation of serialization.
     * Must call HandleDelegate.writeEJBObject, as specified in 19.5.5.1 of spec EJB 2.1
     * @param out The output stream used to write object
     * @throws IOException error when writing object.
     */
    private void writeObject(java.io.ObjectOutputStream out)
    throws IOException {
        HandleDelegate hdld;
        try {
            hdld = JHandleDelegate.getHandleDelegate();
        } catch (NamingException e) {
            throw new IOException("Cannot get HandleDelegate");
        }
        hdld.writeEJBObject(getEJBObject(), out);
    }

    /**
     * Specific implementation of deserialization.
     * Must call HandleDelegate.readEJBObject, as specified in 19.5.5.1 of spec EJB 2.1
     * @param in The input Stream from where is read the object.
     * @throws IOException error when reading object.
     * @throws ClassNotFoundException -
     */
    private void readObject(java.io.ObjectInputStream in)
    throws IOException, ClassNotFoundException {
        HandleDelegate hdld;
        try {
            hdld = JHandleDelegate.getHandleDelegate();
        } catch (NamingException e) {
            throw new IOException("Cannot get HandleDelegate");
        }
        remote = hdld.readEJBObject(in);
        pk = (Serializable) remote.getPrimaryKey();
    }

}

