/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JHomeHandle.java,v 1.8 2004/11/09 16:26:13 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.io.IOException;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Hashtable;

import javax.ejb.EJBHome;
import javax.ejb.HomeHandle;
import javax.ejb.spi.HandleDelegate;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;

import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class implements javax.ejb.HomeHandle interface. A handle is an
 * abstraction of a network reference to a home object. A handle is intended to
 * be used as a "robust" persistent reference to a home object.
 * @author Philippe Durieux, Philippe Coq
 */
public class JHomeHandle implements HomeHandle, Serializable {

    private String homename = null;

    
    private EJBHome ejbHome = null; 

    /**
     * constructor
     * @param hname JNDI name of the Home
     */
    public JHomeHandle(String hname) {
        homename = hname;
    }

    // -----------------------------------------------------------------------
    // HomeHandle implementation
    // -----------------------------------------------------------------------

    /**
     * Obtains the home object represented by this handle.
     * @throws RemoteException The home object could not be obtained because of
     *         a system-level failure.
     * @return The EJBHome object
     */
    public EJBHome getEJBHome() throws RemoteException {
        TraceEjb.interp.log(BasicLevel.DEBUG, homename);
        if (ejbHome == null) {
            try {
                Context ic = new InitialContext();
                ejbHome = (EJBHome) PortableRemoteObject.narrow(ic.lookup(homename), EJBHome.class);
            } catch (NamingException e) {
                throw new RemoteException("getEJBHome:" + e);
            }
        }
        return ejbHome;
    }
    
    /**
     * Specific implementation of serialization.
     * Must call HandleDelegate.writeEJBObject, as specified in 19.5.5.1 of spec EJB 2.1
     * @param out The output stream used to write object
     * @throws IOException error when writing object.
     */
    private void writeObject(java.io.ObjectOutputStream out)
    throws IOException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        HandleDelegate hdld;
        try {
            hdld = JHandleDelegate.getHandleDelegate();
        } catch (NamingException e) {
            throw new IOException("Cannot get HandleDelegate");
        }
        hdld.writeEJBHome(getEJBHome(), out);
    }

    /**
     * Specific implementation of deserialization.
     * Must call HandleDelegate.readEJBObject, as specified in 19.5.5.1 of spec EJB 2.1
     * @param in The input Stream from where is read the object.
     * @throws IOException error when reading object.
     * @throws ClassNotFoundException -
     */
    private void readObject(java.io.ObjectInputStream in)
    throws IOException, ClassNotFoundException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        HandleDelegate hdld;
        try {
            hdld = JHandleDelegate.getHandleDelegate();
        } catch (NamingException e) {
            throw new IOException("Cannot get HandleDelegate");
        }
        try {
            ejbHome = hdld.readEJBHome(in);
        } catch (ClassNotFoundException e) {
            TraceEjb.interp.log(BasicLevel.ERROR, "Cant read EJBHome:" + e);
            throw e;
        } catch (IOException e) {
            TraceEjb.interp.log(BasicLevel.ERROR, "Cant read EJBHome:" + e);
            throw e;
        }
    }

}

