/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JStatefulFactory.java,v 1.16 2004/11/10 15:51:19 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;

import javax.ejb.EJBException;
import javax.ejb.SessionBean;
import javax.ejb.TimerService;

import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatefulDesc;
import org.objectweb.jonas_ejb.lib.EJBInvocation;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a factory for a Session Stateful Bean.
 * @author Philippe Durieux
 */
public class JStatefulFactory extends JSessionFactory {

    JStatefulPool statefulPool;

    /**
     * constructor
     * @param dd Bean Deployment Descriptor
     * @param cont Container where the bean is defined
     * @param sfp Pool des sessions stateful
     */
    public JStatefulFactory(SessionStatefulDesc dd, JContainer cont, JStatefulPool sfp) {
        super((SessionDesc) dd, cont);
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        isSynchro = javax.ejb.SessionSynchronization.class.isAssignableFrom(beanclass);
        isStateful = true;
        statefulPool = sfp;
    }

    // ---------------------------------------------------------------
    // BeanFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the Instance pool size for this Ejb
     */
    public int getPoolSize() {
        return 0; // TODO
    }

    /**
     * Reduce number of instances in memory
     */
    public void reduceCache() {
    }

    // ---------------------------------------------------------------
    // preInvoke / postInvoke
    // ---------------------------------------------------------------

    /**
     * preInvoke for Session beans stateful
     * @param txa Transaction Attribute (Supports, Required, ...)
     * @return A RequestCtx object
     * @throws EJBException
     */
    public RequestCtx preInvoke(int txa) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        RequestCtx rctx = super.preInvoke(txa);
        return rctx;
    }

    /**
     * Check if the access to the bean is authorized
     * @param ejbInv object containing security signature of the method, args of
     *        method, etc
     */
     public void checkSecurity(EJBInvocation ejbInv) {
         TraceEjb.interp.log(BasicLevel.DEBUG, "");
         super.checkSecurity(ejbInv);
     }

     /**
     * postinvoke
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @throws EJBException
     */
    public void postInvoke(RequestCtx rctx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        super.postInvoke(rctx);
    }

    // ---------------------------------------------------------------
    // other public methods
    // ---------------------------------------------------------------

    /**
     * Obtains the TimerService associated for this Bean
     * @return a JTimerService instance.
     */
    public TimerService getTimerService() {
        throw new EJBException("No TimerService for Stateful Session beans");
    }

    /**
     * Creates a new Session Stateful called back from createEJB
     * @return The Session Switch object
     */
    public JSessionSwitch createNewSession() throws RemoteException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JStatefulSwitch bs = new JStatefulSwitch(this, statefulPool);
        return bs;
    }

    /**
     * get a new session context must call newInstance (EJB specs) => no pool
     * should be used.
     * @return a new Session Context
     */
    public JSessionContext getJContext(JSessionSwitch ss) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JStatefulContext bctx = null;
        try {
            bctx = createNewInstance(ss);
        } catch (Exception e) {
            throw new EJBException("Cannot create a new instance " + e);
        }
        return bctx;
    }

    // ---------------------------------------------------------------
    // private methods
    // ---------------------------------------------------------------

    /**
     * Create a new instance of the bean and its StatefulContext
     */
    private JStatefulContext createNewInstance(JSessionSwitch ss) throws Exception {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        // create the bean instance
        SessionBean bean = null;
        try {
            bean = (SessionBean) beanclass.newInstance();
        } catch (InstantiationException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot instantiate session bean");
            throw e;
        } catch (IllegalAccessException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot instantiate session bean");
            throw e;
        }
        // create a new StatefulContext and bind it to the instance
        JStatefulContext bctx = new JStatefulContext(this, bean, isSynchro);
        bean.setSessionContext(bctx);
        bctx.setState(1);
        return bctx;
    }

}
