/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JStatefulSwitch.java,v 1.17 2004/11/26 13:47:03 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.ejb.EJBException;
import javax.ejb.NoSuchObjectLocalException;
import javax.ejb.RemoveException;
import javax.ejb.SessionSynchronization;
import javax.ejb.TransactionRolledbackLocalException;
import javax.transaction.InvalidTransactionException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

import org.objectweb.transaction.jta.ResourceManagerEvent;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JStatefulSwitch is the implementation of JSessionSwitch dedicated to the
 * Stateful Session Bean.
 * @author Philippe Durieux
 */
public class JStatefulSwitch extends JSessionSwitch {

    private int sessionId;

    private JStatefulContext bctx = null;

    private Transaction currTx = null;

    private JStatefulPool statefulPool;

    private boolean mustCommit = false;

    private boolean expired = false;

    private Transaction beanTx = null; // only for bean managed tx

    /**
     * Saved connectionList for this instance. The real connectionList
     * is maintained in a ThreadLocal variable in ThreadData.
     * Used only for session stateful methods, if they keep
     * connection along several calls, not always in same thread.
     * This list must not be shared between all instances.
     * This info is transient because JDBC connection have not to be
     * saved on disk at passivation time : They should be closed in
     * ejbPassivate().
     */
    private List connectionList = Collections.synchronizedList(new ArrayList());

    /**
     * constructor.
     * @param bf The Bean Factory
     * @param The Session Stateful Pool
     */
    public JStatefulSwitch(JStatefulFactory bf, JStatefulPool sfp) throws RemoteException {
        super(bf);
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        statefulPool = sfp;
    }

    /**
     * Set the connection list associated to the current thread
     * with the list associated to this stateful session.
     */
    public void pushConnectionList() {
        if (TraceEjb.isDebugTx()) {
            TraceEjb.tx.log(BasicLevel.DEBUG, "pushed connectionList =" + connectionList);
        }
        bf.getTransactionManager().pushThreadLocalRMEventList(connectionList);
    }

    /**
     * save the current connectionList for future use (next preInvoke).
     */
    public void popConnectionList() {
        connectionList = bf.getTransactionManager().popThreadLocalRMEventList();
        if (TraceEjb.isDebugTx()) {
            TraceEjb.tx.log(BasicLevel.DEBUG, "poped connectionList =" + connectionList);
        }
    }

    /**
     * Save the Connection List after a create method.
     */
    public void setConnectionList(List cl) {
        connectionList = cl;
        if (TraceEjb.isDebugTx()) {
            TraceEjb.tx.log(BasicLevel.DEBUG, "init connectionList =" + connectionList);
        }
    }

    /**
     * enlist all connection of the list
     */
    public void enlistConnections(Transaction tx) {
        if (tx != null && connectionList != null) {
            try {
                for (Iterator it = connectionList.iterator(); it.hasNext();) {
                    ResourceManagerEvent rme = (ResourceManagerEvent) it.next();
                    if (rme != null) {
                        rme.enlistConnection(tx);
                    } else {
                        TraceEjb.tx.log(BasicLevel.WARN, "Null ResourceManagerEvent in Connection List");
                    }
                }
            } catch (SystemException e) {
                TraceEjb.tx.log(BasicLevel.ERROR, "cannot enlist connection" + e);
            }
        }
        // push the connection list in case a tx is started by the bean
        // or if the method close a connection previously enlisted.
        pushConnectionList();
    }

    /**
     * delist all connections of the list
     */
    public void delistConnections(Transaction tx) {
        popConnectionList();
    }

    // ===============================================================
    // TimerEventListener implementation
    // ===============================================================

    /**
     * The session timeout has expired
     * @param arg Not Used.
     */
    public synchronized void timeoutExpired(Object arg) {
        TraceEjb.logger.log(BasicLevel.WARN, "stateful session timeout expired");
        mytimer = null;
        // Do not remove if still used in a transaction
        if (currTx != null) {
            expired = true;
        } else {
            if (bctx != null) {
                try {
                    bctx.setRemoved();
                } catch (RemoteException e) {
                    TraceEjb.logger.log(BasicLevel.WARN, "timeout expired:" + e);
                } catch (RemoveException e) {
                    TraceEjb.logger.log(BasicLevel.WARN, "timeout expired:" + e);
                }
            }
            noLongerUsed();
        }
    }

    // ===============================================================
    // other public methods
    // ===============================================================

    /**
     * At each business method, get a BeanContext to run it
     * @param tx The Transaction Context
     * @return The Session Context
     */
    public JSessionContext getICtx(Transaction tx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // If session has been removed, we must throw
        // NoSuchObject[Local]Exception
        // to the caller.
        if (sessionId == -1) {
            throw new NoSuchObjectLocalException("This Session has been removed");
        }

        // reload Context if it was passivated
        if (bctx == null) {
            bctx = statefulPool.getContext(sessionId);
        }
        // Check Transaction
        checkTx(tx);

        return bctx;
    }

    /**
     * At each create, bind the Context to the transaction
     * @param tx The current Transaction Context
     * @param bctx The Context to bind
     */
    public void bindICtx(Transaction tx, JStatefulContext bctx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        // register this session to the session pool and get an Id
        sessionId = statefulPool.addContext(bctx);
        this.bctx = bctx;
        bctx.initSessionContext(this);
        // Check Transaction
        checkTx(tx);
    }

    /**
     * Release the Context after use.
     * @param tx The current Transaction Context
     */
    public void releaseICtx(Transaction tx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // In case getICtx failed, bctx may be null.
        if (bctx == null) {
            return;
        }

        if (bctx.isMarkedRemoved()) {
            stopTimer();
            noLongerUsed();
        }
    }

    /**
     * Discard a context/instance at end of request. A problem occured on this
     * instance and it must be discarded.
     * @param tx - transaction associated to this context
     */
    public void discardICtx(Transaction tx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // In case getICtx failed, bctx may be null.
        if (bctx == null) {
            return;
        }

        stopTimer();
        noLongerUsed();
    }

    /**
     * This Session is no longer used: - unexport Remote Object - return the
     * Session in the pool
     */
    public void noLongerUsed() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // Unexport the EJBObject from the Orb
        if (remote != null) {
            try {
                remote.unexportObject();
            } catch (NoSuchObjectException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "unexportObject failed: " + e);
            }
        }
        
        // Forget transaction that could be uncommitted.
        // Avoids to get it in another session bean later.
        if (beanTx != null) {
            TraceEjb.tx.log(BasicLevel.WARN, "transaction not ended. forget it");
            beanTx = null;
        }

        // return the SessionSwitch in the pool.
        // will be reused for another Session.
        bf.removeEJB(this);

        // Remove this object.
        // Stateful session contexts are not pooled. (EJB spec.)
        statefulPool.removeContext(sessionId);
        bctx = null;
        sessionId = -1;
    }

    /**
     * End of Transaction
     */
    public void txCompleted() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        currTx = null;
        if (expired) {
            timeoutExpired(null); // try again
        }
    }

    /**
     * This is used for remove on stateful session beans only.
     * @return True if bean is participating in a client transaction
     */
    public boolean isInTransaction() {
        return (currTx != null && !mustCommit);
    }

    /**
     * set a flag to remember that the transaction must be committed
     */
    public void setMustCommit(boolean mc) {
        mustCommit = mc;
    }

    /**
     * Keep the bean opened transaction for later use in other methods. Stateful
     * session bean may open a transaction and use it in other methods. This is
     * called at postInvoke
     */
    public void saveBeanTx() {
        if (bf.isTxBeanManaged()) {
            TraceEjb.tx.log(BasicLevel.DEBUG, "");
            try {
                beanTx = bf.getTransactionManager().suspend();
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot suspend transaction:" + e);
            }
        }
    }    
    
    // ===============================================================
    // private methods
    // ===============================================================

    /**
     * Check Transaction This is called at preInvoke
     * @param tx The current Transaction Context
     * @throws EJBException
     * @throws TransactionRolledbackLocalException
     */
    private void checkTx(Transaction tx) {

        // No check if no Synchro, except for bean managed transaction.
        if (bf.isSessionSynchro() == false) {
            // Resume bean associated transaction.
            // Stateful session bean may open a transaction and use it in other
            // methods.
            if (bf.isTxBeanManaged() && beanTx != null) {
                TraceEjb.tx.log(BasicLevel.DEBUG, "resuming Bean Managed Tx");
                try {
                    bf.getTransactionManager().resume(beanTx);
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot resume transaction:" + e);
                } catch (InvalidTransactionException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "Cannot resume transaction:" + e);
                }
            } else {
                TraceEjb.tx.log(BasicLevel.DEBUG, "no checkTx");
            }
            return;
        }

        if (tx == null) {
            TraceEjb.tx.log(BasicLevel.DEBUG, "(No Tx)");
            if (currTx != null) {
                // A synchronized session must be called in the same transaction
                // between afterBegin and beforeCompletion calls
                TraceEjb.logger.log(BasicLevel.ERROR, "synchronized session called outside transaction context");
                throw new EJBException("Synchronized session called outside transaction context");
            }
        } else {
            TraceEjb.tx.log(BasicLevel.DEBUG, "");
            if (currTx == null) {
                // A new transaction starts on this synchronized session
                try {
                    tx.registerSynchronization(bctx);
                    SessionSynchronization ssbean = (SessionSynchronization) bctx.getInstance();
                    ssbean.afterBegin();
                } catch (RollbackException e) {
                    throw new TransactionRolledbackLocalException("Session rolled back");
                } catch (SystemException e) {
                    throw new EJBException("checkTx error:" + e);
                } catch (RemoteException e) {
                    throw new EJBException("checkTx error:" + e);
                }
                currTx = tx;
            } else {
                // A synchronized session must be called in the same transaction
                // between afterBegin and beforeCompletion calls
                if (tx.equals(currTx) == false) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "synchronized session called in another transaction context");
                    throw new EJBException("Synchronized session called in another transaction context");
                }
            }
        }
    }
}
