/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id: FileManager.java,v 1.2 2004/07/16 13:57:50 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_lib.deployment.work;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;

import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * JOnAS File manager. This class provides a way for managing the working
 * package files.
 * @author Florent Benoit
 * @author Ludovic Bert
 * @author Nicolas Van Caneghem <nicolas.vancaneghem@openpricer.com>Allow the
 *         deployment of an exploded ear
 */

public class FileManager {

    /**
     * Format of a timestamp dir.
     */
    private static final String TIMESTAMP_FORMAT = "_yyyy.MM.dd-HH.mm.ss";

    /**
     * Size of the buffer.
     */
    private static final int BUFFER_SIZE = 2048;

    /**
     * Constructor. Private as it is an utility class
     */
    protected FileManager() {

    }

    /**
     * Give the unpack destination directory of the specified file.
     * @param urlFileName the url of the name of the EAR file (ends with the
     *        .ear extension).
     * @return the timestamp destination directory of the specified file.
     * @throws FileManagerException if we can't get the timestamp.
     */
    public static String fileToTimeStampDir(URL urlFileName) throws FileManagerException {
        return fileToTimeStampDir(urlFileName, "");
    }

    /**
     * Give the unpack destination directory of the specified file.
     * @param urlFileName the url of the name of the EAR file (ends with the
     *        .ear extension).
     * @param ext the suffix to concatenate to the build name
     * @return the timestamp destination directory of the specified file.
     * @throws FileManagerException if we can't get the timestamp.
     */
    public static String fileToTimeStampDir(URL urlFileName, String ext) throws FileManagerException {

        //check protocol
        if (!urlFileName.getProtocol().equalsIgnoreCase("file")) {
            throw new FileManagerException("Only the file:/ URL can be used");
        }

        File urlFile = null;
        urlFile = new File(urlFileName.getFile());

        if (!urlFile.exists()) {
            throw new FileManagerException("The file " + urlFileName.getFile() + " was not found.");
        }

        //Create a format
        SimpleDateFormat sdf = new SimpleDateFormat(TIMESTAMP_FORMAT);

        //get the date of the Ear file
        long lastModified = urlFile.lastModified();

        if (lastModified == 0L) {
            throw new FileManagerException("Can't read the last modified time for the file" + urlFile + ".");
        }

        //Date
        Date d = new Date(lastModified);

        //Return String
        String stReturn = urlFile.getName();
        //Remove extension .ear
        int lastIndex = stReturn.lastIndexOf(".");
        if (lastIndex == -1) {
            throw new FileManagerException("The specified file " + urlFileName.getFile()
                    + " is not a file with the format XXX.ear.");
        }

        stReturn = stReturn.substring(0, lastIndex);

        stReturn += sdf.format(d) + ext;

        return stReturn;
    }

    /**
     * Write an input stream to a given file.
     * @param in the inputStream.
     * @param earEntryFile the file where the inputstream must be dumped.
     * @throws FileManagerException if the dump failed.
     */
    protected static void dump(InputStream in, File earEntryFile) throws FileManagerException {

        try {
            //File output
            FileOutputStream out = new FileOutputStream(earEntryFile);
            int n = 0;
            try {
                //buffer
                byte[] buffer = new byte[BUFFER_SIZE];

                while ((n = in.read(buffer)) > 0) {
                    out.write(buffer, 0, n);
                }
            } finally {
                out.close();
            }
        } catch (IOException e) {
            throw new FileManagerException("Error while uncompressing the file " + earEntryFile + ": " + e.getMessage());
        }
    }

    /**
     * Copy a file
     * @param pSrcFilename source file
     * @param pDstFilename dest file
     * @throws FileManagerException if the copy of the file failed
     */
    public static void copyFile(String pSrcFilename, String pDstFilename) throws FileManagerException {
        try {

            // input
            File sourceFile = new File(pSrcFilename);
            FileInputStream fis = new FileInputStream(sourceFile);
            BufferedInputStream bis = new BufferedInputStream(fis);
            long l = sourceFile.length();

            // ouput
            FileOutputStream fos = new FileOutputStream(pDstFilename);
            BufferedOutputStream bos = new BufferedOutputStream(fos);

            // Copy
            for (long i = 0; i < l; i++) {
                bos.write(bis.read());
            }

            // Close streams
            bos.flush();
            bos.close();
            bis.close();

        } catch (Exception e) {

            throw new FileManagerException("Error during copy file : " + pSrcFilename + " -> " + pDstFilename);

        }

    }
}