/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2003-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: Config.java,v 1.1 2004/10/11 13:16:14 benoitf Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_lib.genbase.generator;

import java.io.File;
import java.util.Vector;

import org.objectweb.common.Env;

import org.objectweb.jonas.server.JClassLoader;


/**
 * Configuration object storing generation params.
 */
public class Config {

    /** packed mode */
    public static final int PACKED = 0;

    /** unpacked mode */
    public static final int UNPACKED = 1;

    /** Java executable name */
    private String nameJava = "java";

    /** Javac executable name */
    private String nameJavac = "javac";

    /**
     * Rmic executable name
     */
    private String nameRmic = "rmic";

    /** Javac options */
    private Vector javacOpts = new Vector();

    /** Bin directory for JDK */
    private String javaHomeBin = null;

    /** no config needed */
    private boolean noConfig = false;

    /** verbose mode */
    private boolean verbose = false;

    /** debug mode */
    private boolean debug = false;

    /** classpath */
    private String classpath = ".";

    /** output directory */
    private File out = new File(".");

    /** keep alrady generated files */
    private boolean keepGenerated = false;

    /** validating parser ? */
    private boolean parseWithValidation = true;

    /** configuration error */
    private boolean error = false;

    /** file input name */
    private String inputname;

    /** help requested ? */
    private boolean help = false;

    /** save mode (PACKED/UNPACKED) */
    private int saveMode = PACKED;

    /**
     * Creates a new Config. Automatically setup javaHomeBin property. And
     * create classpath from classloader.
     */
    public Config() {
        // Setup java_home/bin directory
        setJavaHomeBin(System.getProperty("java.home", ""));

        if (!("".equals(getJavaHomeBin()))) {
            if (Env.isOsMacOsX()) {
                setJavaHomeBin(getJavaHomeBin() + File.separator + "bin" + File.separator);
            } else {
                // JRE Directory !
                setJavaHomeBin(getJavaHomeBin() + File.separator + ".." + File.separator + "bin" + File.separator);
            }
        }

        // classpath
        JClassLoader ucl = (JClassLoader) (Thread.currentThread().getContextClassLoader());
        setClasspath(ucl.getClassPath());
    }

    /**
     * @param nameJavac Set javac command name to use
     */
    public void setNameJavac(String nameJavac) {
        this.nameJavac = nameJavac;
    }

    /**
     * @return Returns the javac command name to use
     */
    public String getNameJavac() {
        return nameJavac;
    }

    /**
     * @param javacOpts Set javac Opts
     */
    public void setJavacOpts(Vector javacOpts) {
        this.javacOpts = javacOpts;
    }

    /**
     * @return Returns the Javac Opts to use
     */
    public Vector getJavacOpts() {
        return javacOpts;
    }

    /**
     * @param javaHomeBin JAVA_HOME/bin directory
     */
    public void setJavaHomeBin(String javaHomeBin) {
        this.javaHomeBin = javaHomeBin;
    }

    /**
     * @return Returns the JAVA_HOME/bin directory
     */
    public String getJavaHomeBin() {
        return javaHomeBin;
    }

    /**
     * @param noConfig Generate Configuration ?
     */
    public void setNoConfig(boolean noConfig) {
        this.noConfig = noConfig;
    }

    /**
     * @return Returns noConfig option
     */
    public boolean isNoConfig() {
        return noConfig;
    }

    /**
     * @param verbose Verbose ?
     */
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    /**
     * @return Returns verbose option
     */
    public boolean isVerbose() {
        return verbose;
    }

    /**
     * @param debug Debug ?
     */
    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    /**
     * @return Returns debug option
     */
    public boolean isDebug() {
        return debug;
    }

    /**
     * @param classpath Classpath to use with java commands
     */
    public void setClasspath(String classpath) {
        this.classpath = classpath;
    }

    /**
     * @return Retruns Classpath
     */
    public String getClasspath() {
        return classpath;
    }

    /**
     * @param out Output directory
     */
    public void setOut(File out) {
        this.out = out;
    }

    /**
     * @return Returns Ouput directory
     */
    public File getOut() {
        return out;
    }

    /**
     * @param keepGenerated Kepp Generated files ?
     */
    public void setKeepGenerated(boolean keepGenerated) {
        this.keepGenerated = keepGenerated;
    }

    /**
     * @return Returns keepGenerated option
     */
    public boolean isKeepGenerated() {
        return keepGenerated;
    }

    /**
     * @param parseWithValidation Parse XML desc with validation ?
     */
    public void setParseWithValidation(boolean parseWithValidation) {
        this.parseWithValidation = parseWithValidation;
    }

    /**
     * @return Returns validation
     */
    public boolean isParseWithValidation() {
        return parseWithValidation;
    }

    /**
     * @param error Error Mode ?
     */
    public void setError(boolean error) {
        this.error = error;
    }

    /**
     * @return Returns true if there is Configuration errors
     */
    public boolean isError() {
        return error;
    }

    /**
     * @param inputname File inputname
     */
    public void setInputname(String inputname) {
        this.inputname = inputname;
    }

    /**
     * @return Returns file input name
     */
    public String getInputname() {
        return inputname;
    }

    /**
     * @param help Help Mode ?
     */
    public void setHelp(boolean help) {
        this.help = help;
    }

    /**
     * @return Returns Help option.
     */
    public boolean isHelp() {
        return help;
    }

    /**
     * Set Packed Mode for storing.
     */
    public void setSavePacked() {
        this.saveMode = PACKED;
    }

    /**
     * Set UnPacked Mode for storing.
     */
    public void setSaveUnpacked() {
        this.saveMode = UNPACKED;
    }

    /**
     * @return Returns Save mode
     */
    public int getSaveMode() {
        return saveMode;
    }
    /**
     * @return Returns the nameRmic.
     */
    public String getNameRmic() {
        return nameRmic;
    }
    /**
     * @param nameRmic The nameRmic to set.
     */
    public void setNameRmic(String nameRmic) {
        this.nameRmic = nameRmic;
    }
    /**
     * @return the java command
     */
    public String getNameJava() {
        return nameJava;
    }
}