/**
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: FileUtils.java,v 1.1 2004/10/11 13:16:14 benoitf Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_lib.genbase.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/**
 * <code>FileUtils</code> is an utility class with no instance. Used to pack a
 * directory into a jar archive, copy a file, move a file, ...
 *
 * @author Guillaume Sauthier
 */
public class FileUtils {

    /** Buffer Size */
    public static final int MAX_BUFFER_SIZE = 1024;

    /**
     * Empty Constructor for Utility Class
     */
    private FileUtils() {
    }

    /**
     * Unpack the source archive in a temporary directory and returns temporary
     * directory created.
     *
     * @param source source JarFile to be unpacked
     *
     * @return root directory containing unpacked archive.
     *
     * @throws IOException When unpack fails
     */
    public static File unpack(JarFile source) throws IOException {
        TempRepository tr = TempRepository.getInstance();
        File tmp = tr.createDir();

        for (Enumeration e = source.entries(); e.hasMoreElements();) {
            ZipEntry ze = (ZipEntry) e.nextElement();

            if (!ze.getName().endsWith("/")) {
                unpack(ze.getName(), source.getInputStream(ze), tmp);
            }
        }

        return tmp;
    }

    /**
     * Unpack a jar entry (name + is) is root directory (crate directory if
     * needed)
     *
     * @param name entry name
     * @param is file inputStream
     * @param root root directory
     *
     * @throws IOException When unpack fails
     */
    private static void unpack(String name, InputStream is, File root) throws IOException {
        // create directories if needed
        String sysfilename = name.replace('/', File.separator.toCharArray()[0]);
        String filename = sysfilename.substring(sysfilename.lastIndexOf(File.separator) + 1);
        String dirname = sysfilename.substring(0, sysfilename.lastIndexOf(File.separator) + 1);
        File dir = new File(root, dirname);

        if (!dir.exists()) {
            dir.mkdirs();
        }

        // fill the new file
        OutputStream os = new FileOutputStream(new File(dir, filename));

        fill(is, os);

        os.close();
        is.close();
    }

    /**
     * Fill an OutputStream with the content of an InputStream.
     *
     * @param is input content
     * @param os output
     *
     * @throws IOException When filling cannot complete
     */
    private static void fill(InputStream is, OutputStream os) throws IOException {
        byte[] buffer = new byte[MAX_BUFFER_SIZE];
        int read;

        while ((read = is.read(buffer, 0, MAX_BUFFER_SIZE)) != -1) {
            os.write(buffer, 0, read);
        }
    }

    /**
     * Copy a source file into a destination directory.
     *
     * @param source source file
     * @param target destination directory
     *
     * @throws IOException When copy cannot succeed
     */
    public static void copy(File source, File target) throws IOException {
        InputStream is = new FileInputStream(source);
        OutputStream os = new FileOutputStream(new File(target, source.getName()));
        fill(is, os);
        is.close();
        os.close();

    }
}