/**
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: WsEndpointDDModifier.java,v 1.4 2004/04/19 13:49:54 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.ddmodifier;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Modify a Deployment Desc for Endpoint. Wrapper around a web.xml DOM.
 *
 * @author Guillaume Sauthier
 */
public class WsEndpointDDModifier extends DeploymentDescModifier {

    /**
     * servlet Element name
     */
    private static final String SERVLET = "servlet";

    /**
     * servlet-name Element name
     */
    private static final String SERVLET_NAME = "servlet-name";

    /**
     * servlet-class Element name
     */
    private static final String SERVLET_CLASS = "servlet-class";

    /**
     * servlet-mapping Element name
     */
    private static final String SERVLET_MAPPING = "servlet-mapping";

    /**
     * url-pattern Element name
     */
    private static final String URL_MAPPING = "url-pattern";

    /**
     * init-param Element name
     */
    private static final String INIT_PARAM = "init-param";

    /**
     * param-name Element name
     */
    private static final String PARAM_NAME = "param-name";

    /**
     * param-value Element name
     */
    private static final String PARAM_VALUE = "param-value";

    /**
     * Creates a new WsEndpointDDModifier object.
     *
     * @param web web.xml document
     */
    public WsEndpointDDModifier(Document web) {
        super(web.getDocumentElement(), web);
    }

    /**
     * Add a new <code>servlet</code> element in the web.xml.
     *
     * @param name servlet name.
     * @param classname servlet fully qualified classname.
     */
    public void addServlet(String name, String classname) {
        Element servlet = newJ2EEElement(SERVLET);
        Element servletName = newJ2EEElement(SERVLET_NAME, name);
        Element servletClass = newJ2EEElement(SERVLET_CLASS, classname);

        servlet.appendChild(servletName);
        servlet.appendChild(servletClass);

        // add servlet in the webapp Element
        getElement().appendChild(servlet);
    }

    /**
     * Remove a <code>servlet</code> element from the web.xml.
     *
     * @param name servlet name.
     */
    public void removeServlet(String name) {
        Element servlet = findServlet(name);

        if (servlet != null) {
            getElement().removeChild(servlet);
        }
    }

    /**
     * Add a new <code>servlet-mapping</code> element in the web.xml.
     *
     * @param name servlet name.
     * @param mapping <code>url-mapping</code> value
     */
    public void addServletMapping(String name, String mapping) {
        Element servletMapping = newJ2EEElement(SERVLET_MAPPING);
        Element servletName = newJ2EEElement(SERVLET_NAME, name);
        Element urlMapping = newJ2EEElement(URL_MAPPING, mapping);

        servletMapping.appendChild(servletName);
        servletMapping.appendChild(urlMapping);

        // add servletMapping in the webapp Element
        getElement().appendChild(servletMapping);
    }

    /**
     * Add a new <code>init-param</code> element in the web.xml.
     *
     * @param servletName the servlet name where init-param will be added.
     * @param pName parameter name
     * @param pValue parameter value
     */
    public void addServletParam(String servletName, String pName, String pValue) {
        Element ip = newJ2EEElement(INIT_PARAM);
        Element pn = newJ2EEElement(PARAM_NAME, pName);
        Element pv = newJ2EEElement(PARAM_VALUE, pValue);

        ip.appendChild(pn);
        ip.appendChild(pv);

        Element servlet = findServlet(servletName);
        servlet.appendChild(ip);

    }

    /**
     * search webapp element for servlet named with the given name.
     *
     * @param name the searched servlet name
     *
     * @return the found element or null if element is not found (should'nt
     *         occurs).
     */
    private Element findServlet(String name) {
        NodeList nl = getElement().getElementsByTagNameNS(J2EE_NS, SERVLET);
        Element servlet = null;

        for (int i = 0; (i < nl.getLength()) && (servlet == null); i++) {
            Element e = (Element) nl.item(i);

            NodeList names = e.getElementsByTagNameNS(J2EE_NS, SERVLET_NAME);

            // test servlet/servlet-name/#text-node.value
            if (names.item(0).getFirstChild().getNodeValue().equals(name)) {
                servlet = e;
            }
        }

        return servlet;
    }

    /**
     * Remove servlet-mapping tag associated to a given servlet-name
     *
     * @param sName servlet-name
     */
    public void removeServletMapping(String sName) {
        NodeList nl = getElement().getElementsByTagNameNS(J2EE_NS, SERVLET_MAPPING);
        Element mapping = null;

        for (int i = 0; (i < nl.getLength()) && (mapping == null); i++) {
            Element e = (Element) nl.item(i);

            NodeList names = e.getElementsByTagNameNS(J2EE_NS, SERVLET_NAME);

            // test servlet-mapping/servlet-name/#text-node.value
            if (names.item(0).getFirstChild().getNodeValue().equals(sName)) {
                mapping = e;
            }
        }

        if (mapping != null) {
            getLogger().log(BasicLevel.DEBUG, "mapping element found : " + mapping);
            getElement().removeChild(mapping);
        }

    }

    /**
     * DOCUMENT ME!
     *
     * @param name DOCUMENT ME!
     * @param home DOCUMENT ME!
     * @param remote DOCUMENT ME!
     * @param link DOCUMENT ME!
     */
    public void addEjbRef(String name, String home, String remote, String link) {

        Element ejbRef = newJ2EEElement("ejb-ref");
        Element ejbRefName = newJ2EEElement("ejb-ref-name", name);
        Element ejbRefType = newJ2EEElement("ejb-ref-type", "Session");
        Element ejbHome = newJ2EEElement("home", home);
        Element ejbRemote = newJ2EEElement("remote", remote);
        Element ejbLink = newJ2EEElement("ejb-link", link);

        ejbRef.appendChild(ejbRefName);
        ejbRef.appendChild(ejbRefType);
        ejbRef.appendChild(ejbHome);
        ejbRef.appendChild(ejbRemote);
        ejbRef.appendChild(ejbLink);

        getElement().appendChild(ejbRef);

    }

    /**
     * DOCUMENT ME!
     *
     * @param name DOCUMENT ME!
     * @param home DOCUMENT ME!
     * @param remote DOCUMENT ME!
     * @param link DOCUMENT ME!
     */
    public void addEjbLocalRef(String name, String home, String remote, String link) {

        Element ejbRef = newJ2EEElement("ejb-local-ref");
        Element ejbRefName = newJ2EEElement("ejb-ref-name", name);
        Element ejbRefType = newJ2EEElement("ejb-ref-type", "Session");
        Element ejbHome = newJ2EEElement("local-home", home);
        Element ejbRemote = newJ2EEElement("local", remote);
        Element ejbLink = newJ2EEElement("ejb-link", link);

        ejbRef.appendChild(ejbRefName);
        ejbRef.appendChild(ejbRefType);
        ejbRef.appendChild(ejbHome);
        ejbRef.appendChild(ejbRemote);
        ejbRef.appendChild(ejbLink);

        getElement().appendChild(ejbRef);

    }
}