/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2003-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: GeneratorFactory.java,v 1.6 2004/10/11 13:16:14 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.generator;

import org.objectweb.jonas_lib.I18n;
import org.objectweb.jonas_lib.genbase.GenBaseException;
import org.objectweb.jonas_lib.genbase.archive.Archive;
import org.objectweb.jonas_lib.genbase.generator.Config;

import org.objectweb.jonas_ws.deployment.api.ServiceDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceRefDesc;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.ddmodifier.WebServicesDDModifier;
import org.objectweb.jonas_ws.wsgen.ddmodifier.WsClientDDModifier;
import org.objectweb.jonas_ws.wsgen.ddmodifier.WsEndpointDDModifier;

import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * a <code>GeneratorFactory</code> has to be extended by specific generation
 * mecanism. It will look in <code>jonas.properties</code> file for a property
 * named :<code>jonas.service.ws.wsgen.generator.factory</code> that is a
 * classname extending <code>GeneratorFactory</code>. By default Axis
 * GeneratorFactory is used.
 *
 * @author Guillaume Sauthier
 */
public abstract class GeneratorFactory {

    /**
     * Generator Factory property name in jonas.properties
     */
    public static final String GENERATOR_FACTORY = "jonas.service.ws.wsgen.generator.factory";

    /**
     * Default GeneratorFactory impl to use
     */
    public static final String GENERATOR_FACTORY_DEFAULT = "org.objectweb.jonas_ws.wsgen.generator.axis.AxisGeneratorFactory";

    /** <code>GeneratorFactory</code> unique instance */
    private static GeneratorFactory instance = null;

    /** i18n */
    private static I18n i18n = I18n.getInstance(GeneratorFactory.class);

    /** logger */
    private static Logger logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);

    /** Configuration to set on instanciated Generator */
    private Config configuration;

    /**
     * Returns the unique GeneratorFactory instance.
     *
     * @return the unique GeneratorFactory instance.
     *
     * @throws WsGenException When instanciation fails
     */
    public static GeneratorFactory getInstance() throws WsGenException {
        if (instance == null) {
            instance = newInstance();
        }

        return instance;
    }

    /**
     * Create a new generatorFactory instance by looking up in
     * <code>jonas.properties</code> and load class specified with
     * <code>jonas.service.ws.wsgen.generator-factory</code> property. If not
     * set, Axis GeneratorFactory is the default Factory returned.
     *
     * @return a new generatorFactory instance.
     *
     * @throws WsGenException when Exception occurs when dynamiccaly
     *         instantiating GeneratorFactory subclass
     */
    private static GeneratorFactory newInstance() throws WsGenException {
        // Lookup property
        JProp jp;
        String classname = null;

        try {
            jp = JProp.getInstance();
            classname = jp.getValue(GENERATOR_FACTORY, GENERATOR_FACTORY_DEFAULT);
        } catch (Exception e) {
            // do nothing, just log a warning a continue with default values...
            logger.log(BasicLevel.WARN, "Cannot get '" + GENERATOR_FACTORY + "' value, default used : '"
                    + GENERATOR_FACTORY_DEFAULT + "'");
            classname = GENERATOR_FACTORY_DEFAULT;
        }

        // instanciate
        GeneratorFactory gf = null;

        try {
            ClassLoader jcl = Thread.currentThread().getContextClassLoader();
            Class gfc = jcl.loadClass(classname);
            gf = (GeneratorFactory) gfc.newInstance();
        } catch (Exception e) {
            String err = i18n.getMessage("GeneratorFactory.newInstance.instance", classname);
            throw new WsGenException(err, e);
        }

        return gf;

    }

    /**
     * Return a new WsClientGenerator for the specific generation mecanism.
     *
     * @param serviceRef the service-ref containing information for client side
     *        generation process.
     * @param ddm the XML modifier.
     * @param archive the Archive to be modified
     *
     * @return a new WsClientGenerator.
     *
     * @throws GenBaseException When Factory cannot instanciate WsClientGenerator
     */
    public abstract WsClientGenerator newGenerator(ServiceRefDesc serviceRef, WsClientDDModifier ddm, Archive archive)
            throws GenBaseException;

    /**
     * Return a new WsEndpointGenerator for the specific generation mecanism.
     *
     * @param serviceDesc the webservice-description containing information for
     *        server side generation process.
     * @param ddm the XML modifier.
     * @param wsddm the Webservices.xml DD modifier
     *
     * @return a new WsEndpointGenerator.
     *
     * @throws GenBaseException When Factory cannot instanciate WsEndpointGenerator
     */
    public abstract WsEndpointGenerator newGenerator(ServiceDesc serviceDesc, WsEndpointDDModifier ddm,
            WebServicesDDModifier wsddm) throws GenBaseException;

    /**
     * Set the Configuration to use with newly created Generator.
     *
     * @param config the Configuration to use with newly created Generator.
     */
    public void setConfiguration(Config config) {
        this.configuration = config;
    }

    /**
     * Get the Configuration to use with newly created Generator.
     *
     * @return the Configuration to use with newly created Generator
     */
    public Config getConfiguration() {
        return configuration;
    }
}